"""Memory mapped dict based using LMDB.

Adopted from here: https://github.com/Dobatymo/lmdb-python-dbm/tree/master
"""

import os
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Tuple

import lmdb

from .. import file_utils


class LMDBDict:
    """A memory mapped variant of python dictionary with limited functionalities.

    You can use this class to decrease the memory footprint of large python dictionaries.


    ..  warning::

        This is WIP. Use this class with extra causion to avoid logical bugs.

    If the following conditions are not met, it is not guaranteed that this class functions as expected and correctly:

        - Only one worker at a time (across processes, threads, etc.) should write to this container.
        - Do not update or overwrite the database file after its creation. I.e., no write operations after you created the database file.
        - Make sure you instantiate reader instances of this class after its creation is completed.
          Readers that are created before the write operation ends might represent an inconsistent view of the data.

    You can follow these requirements by simply doing the following in your code:

    .. code-block:: python

        with file_utils.easyfilelock(db_file):
            if not Path(db_file).exists():
                LMDBDict.create_db(data=dict(...), db_file=db_file)
        mm_dict = LMDBDict.load_db(db_file=db_file, ...)
        print(mm_dict[b'some_key'])
        mm_dict.close()

    You can also use this class as a context manager to properly manage opening
    and closing LMDB environments.

    .. code-block:: python

        with LMDBDict.load_db(...) as db:
            print(db[b'some_key'])

    ..  note::

        from `LMDB docs <http://www.lmdb.tech/doc/starting.html>`_:

        LMDB uses POSIX locks on files, and these locks have issues if
        one process opens a file multiple times. Because of this, do
        not mdb_env_open() a file multiple times from a single process.
        Instead, share the LMDB environment that has opened the file
        across all threads. Otherwise, if a single process opens the
        same environment multiple times, closing it once will remove
        all the locks held on it, and the other instances will be
        vulnerable to corruption from other processes.

    To make sure, we only create one lmdb environment for each file, we save the path to
    the DB file and its corresponding environment in ``lmdb_envs`` class attribute whenever we
    create a new environment. If the user calls ``LMDBDict.load_db()`` for the same file again,
    we reuse the saved environment.
    """

    lmdb_envs = dict()

    def __init__(
        self,
        env: lmdb.Environment,
        encode_key: Optional[Callable[[Any], bytes]] = None,
        decode_key: Optional[Callable[[bytes], Any]] = None,
        encode_value: Optional[Callable[[Any], bytes]] = None,
        decode_value: Optional[Callable[[bytes], Any]] = None,
    ) -> None:
        """Initialize a read-only memory mapped dict with data saved in the given lmdb environment.

        The remaining arguments are the same as arguments to :py:meth:`trove.containers.lmdb_dict.LMDBDict.load_db`.

        Args:
            env: the LMDB environment that saves the underlying data.
        """
        self.env = env
        self.encode_key = encode_key if encode_key is not None else lambda x: x
        self.decode_key = decode_key if decode_key is not None else lambda x: x
        self.encode_value = encode_value if encode_value is not None else lambda x: x
        self.decode_value = decode_value if decode_value is not None else lambda x: x

    @classmethod
    def load_db(
        cls,
        db_file: os.PathLike,
        db_hash: Optional[str] = None,
        encode_key: Optional[Callable[[Any], bytes]] = None,
        decode_key: Optional[Callable[[bytes], Any]] = None,
        encode_value: Optional[Callable[[Any], bytes]] = None,
        decode_value: Optional[Callable[[bytes], Any]] = None,
    ):
        """Load a memory mapped dict from disk.

        Args:
            db_file: path to lmdb database file that holds the data.
            db_hash: If given, it checks if hash of the given ``db_file`` is equal to ``db_hash``.
                Since we are not properly closing the readonly environments on exit, this optional hash
                comparison ensures that DB files are not corrupted in previous runs.
            encode_key: callable with one positional argument that takes in a dictionary key and encodes into bytestring.
                If not provided, keys must be of type ``bytes`` already.
            decode_key: callable with one positional argument that takes in a bytestring and returns the decoded dict key.
                If not provided, this class returns the saved dict key which is of type ``bytes``.
            encode_value: callable with one positional argument that takes in a dict value and encodes into bytestring.
                If not provided, values must be of type ``bytes`` already.
            decode_value: callable with one positional argument that takes in a bytestring and returns the decoded dict value.
                If not provided, this class returns the saved dict value which is of type ``bytes``.

        Returns:
            an instance of LMDBDict populated with the data from the given db file.
        """
        if not Path(db_file).exists():
            msg = f"DB file does not exists:\n{db_file}"
            raise RuntimeError(msg)

        _real_filepath = file_utils.realpath(db_file)
        if _real_filepath in cls.lmdb_envs:
            # We've already created/opened an environment for this db file in this process.
            # Reuse the previous environment instance. Accoding to LMDB docs, we should create one
            # environment for each db file in each process.
            env = cls.lmdb_envs[_real_filepath]["env"]
            # Keep track of the number of instances that use this environment instance.
            # Useful for a proper cleanup-on-exit procedure in the future
            cls.lmdb_envs[_real_filepath]["num_refs"] += 1
        else:
            if db_hash is not None:
                # If a hash is provided, make sure the db file is not corrupted.
                file_hash = file_utils.hash_file_bytes(path=_real_filepath)
                if file_hash.strip() != db_hash.strip():
                    msg = f"DB file content has changed after it has been created. DB file: '{db_file}'"
                    raise RuntimeError(msg)

            env = lmdb.Environment(
                path=_real_filepath,
                map_size=100 * 2**30,
                subdir=False,
                readonly=True,
                readahead=False,
                max_dbs=0,
            )
            # save a pointer to this environment to potentially create other
            # LMDBDict instances with the same DB file in the future
            cls.lmdb_envs[_real_filepath] = {"env": env, "num_refs": 1}
        obj = cls(
            env=env,
            encode_key=encode_key,
            decode_key=decode_key,
            encode_value=encode_value,
            decode_value=decode_value,
        )
        return obj

    @classmethod
    def create_db(
        cls,
        data: Iterable[Tuple[Any, Any]],
        db_file: os.PathLike,
        encode_key: Optional[Callable[[Any], bytes]] = None,
        encode_value: Optional[Callable[[Any], bytes]] = None,
    ) -> str:
        """Create an LMDB database from the given data.

        To use the created database, you should load it into an instance of LMDBDict:

        .. code-block:: python

            LMDBDict.create_db(data=[(b'a', b'val'), (b'b', b'v2')], db_file='./somefile.db')
            mm_dict = LMDBDict.load_db(db_file='./somefile.db')
            assert mm_dict[b'a'] == b'val'

        Args:
            data: an iterable of (key, value) 2-tuples. E.g., ``[(b'a', b'aa'), (b'c', b'cc')]``.
                We do not verify the data type of keys and values. You must make sure to provide
                keys and values of correct data type (i.e., ``bytes```).
            db_file: Path to the destination DB file.
            encode_key: callable with one positional argument that takes in a dict key and encodes into bytestring.
                If not provided, keys must be of type ``bytes`` already.
            encode_value: callable with one positional argument that takes in a dict value and encodes into bytestring.
                If not provided, values must be of type ``bytes`` already.

        Returns:
            hash of the created database file.
        """
        # just to avoid extra if statements in this function
        if encode_key is None:
            encode_key = lambda x: x
        if encode_value is None:
            encode_value = lambda x: x

        Path(db_file).parent.mkdir(exist_ok=True, parents=True)
        _real_fp = file_utils.realpath(db_file)
        with file_utils.easyfilelock(str(_real_fp) + ".is_writing"):
            # Weak file locking. This only ensures no instance of this function is writing to this file.
            # Other functions can still modify this file. But, that is ok, we ask the user
            # to make sure that no other processes is writing to the db file when calling this function.
            # See class docstrings for more info.
            with file_utils.atomic_write(file=_real_fp, root="parent") as tfp:
                with lmdb.Environment(
                    path=tfp,
                    map_size=100 * 2**30,
                    subdir=False,
                    readonly=False,
                    mode=493,
                    readahead=False,
                    max_dbs=0,
                ) as env:
                    with env.begin(write=True) as txn:
                        with txn.cursor() as curs:
                            curs.putmulti(
                                (encode_key(k), encode_value(v)) for k, v in data
                            )
                        env.sync()
            db_hash = file_utils.hash_file_bytes(path=_real_fp)
        return db_hash

    def close(self) -> None:
        """Close LMDB environment.

        It reduces the number of references to the current environment by one (i.e., soft close).
        If there are no remaining references to this environment, it closes the environment.
        """
        db_path = self.env.path()
        self.__class__.lmdb_envs[db_path]["num_refs"] -= 1
        if self.__class__.lmdb_envs[db_path]["num_refs"] < 1:
            self.__class__.lmdb_envs[db_path]["env"].close()

    def __enter__(self):
        """Act as a context manager."""
        return self

    def __exit__(self, *args, **kwargs):
        """Cleanups before exiting the context."""
        self.close()

    def get(self, key: Any, default: Optional[Any]) -> Any:
        """Similar to python's ``dict.get`` method.

        But key must be of type bytes.
        """
        with self.env.begin() as txn:
            value = txn.get(self.encode_key(key))
        if value is None:
            return default
        else:
            return self.decode_value(value)

    def keys(self):
        """Similar to python's ``dict.keys``."""
        with self.env.begin() as txn:
            with txn.cursor() as curs:
                for k in curs.iternext(keys=True, values=False):
                    yield self.decode_key(k)

    def values(self):
        """Similar to python's ``dict.values``."""
        with self.env.begin() as txn:
            with txn.cursor() as curs:
                for v in curs.iternext(keys=False, values=True):
                    yield self.decode_value(v)

    def items(self):
        """Similar to python's ``dict.items``."""
        with self.env.begin() as txn:
            with txn.cursor() as curs:
                for k, v in curs.iternext(keys=True, values=True):
                    yield (self.decode_key(k), self.decode_value(v))

    def __len__(self) -> int:
        """Similar to python's ``len(dict)``."""
        with self.env.begin() as txn:
            num_entries = txn.stat()["entries"]
        return num_entries

    def __contains__(self, key: Any) -> bool:
        """Support ``in`` operator like ``key in dict``."""
        with self.env.begin() as txn:
            value = txn.get(self.encode_key(key))
        return value is not None

    def __iter__(self):
        """Support iterating directly over LMDBDict instances similar to python's dictionary."""
        return self.keys()

    def __getitem__(self, key: Any) -> Any:
        """support indexing operator."""
        with self.env.begin() as txn:
            value = txn.get(self.encode_key(key))
        if value is None:
            raise KeyError(key)
        return self.decode_value(value)



"""Modified transformers trainer class to support saving custom retriever models."""

import os
from dataclasses import dataclass
from typing import Optional

import torch
import torch.distributed as dist
from transformers import Trainer, TrainingArguments
from transformers.trainer import TRAINING_ARGS_NAME
from transformers.utils import logging

from .evaluation import IRMetrics
from .logging_utils import get_logger_with_config
from .modeling import PretrainedRetriever

logger = logging.get_logger(__name__)


@dataclass
class RetrievalTrainingArguments(TrainingArguments):
    trove_logging_mode: str = "all"
    """Determines which processes can use the logging module. It is just a soft limit:
    the excluded processes can still log messages but their logging level is set to `ERROR`.
    You can choose from one of ``['main', 'local_main', and 'all']`` values.
    """

    def __post_init__(self):
        super().__post_init__()

        # set 'ddp_find_unused_parameters' to False if gradient checkpointing is enabled
        # in transformers.Trainer it is only done for instances of PreTrainedModel
        # The related snippet is here: https://github.com/huggingface/transformers/blob/f5620a76344595dbc7c9cff97bbd1edc1696854d/src/transformers/trainer.py#L2046
        # Since we are not an instance of PreTrainedModel, we should do it ourselves
        if self.gradient_checkpointing:
            # Make sure `ddp_find_unused_parameters` is not set to something that contradicts what we want to do
            # It is to safe guard against future changes in transformers inner workings
            assert self.ddp_find_unused_parameters in [None, False]
            self.ddp_find_unused_parameters = False

        if self.trove_logging_mode not in ["main", "local_main", "all"]:
            msg = "Valid values for mode configs (print and pbar) are ['main', 'local_main', and 'all']"
            raise ValueError(msg)

        if dist.is_initialized():
            world_size = dist.get_world_size()
        else:
            world_size = 1
        if world_size < 2:
            return
        log_level = None
        if (self.trove_logging_mode == "main" and self.process_index != 0) or (
            self.trove_logging_mode == "local_main" and self.local_process_index != 0
        ):
            log_level = "ERROR"
        get_logger_with_config(
            name="trove", log_level=log_level, rank=self.process_index, force=True
        )


class RetrievalTrainer(Trainer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/trainer.py#L18
        self.__is_ddp = dist.is_initialized()
        self._dist_loss_scale_factor = dist.get_world_size() if self.__is_ddp else 1

        # Chech if the passed compute_metrics class supports the given training arguments
        if self.compute_metrics is not None and isinstance(
            self.compute_metrics, IRMetrics
        ):
            self.compute_metrics.check_training_arguments(args=self.args)

        if self.train_dataset is not None and hasattr(
            self.train_dataset, "set_trainer"
        ):
            self.train_dataset.set_trainer(self)

    def training_step(self, *args, **kwargs):
        # https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/trainer.py#L53C66-L53C96
        return super().training_step(*args, **kwargs) / self._dist_loss_scale_factor

    def compute_loss(self, *args, **kwargs):
        super_output = super().compute_loss(*args, **kwargs)
        if "model" in kwargs:
            model = kwargs["model"]
        else:
            model = args[0]
        if model.training:
            if isinstance(super_output, tuple):
                return (super_output[0] * self._dist_loss_scale_factor, super_output[1])
            else:
                return super_output * self._dist_loss_scale_factor
        else:
            return super_output

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.

        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.debug(f"Saving model checkpoint to {output_dir}")

        if isinstance(self.model, PretrainedRetriever):
            # https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/trainer.py#L33
            if state_dict is None:
                state_dict = self.model.state_dict()
            self.model.save_pretrained(
                output_dir,
                state_dict=state_dict,
                safe_serialization=self.args.save_safetensors,
            )
        elif isinstance(self.accelerator.unwrap_model(self.model), PretrainedRetriever):
            if state_dict is None:
                state_dict = self.model.state_dict()
            self.accelerator.unwrap_model(self.model).save_pretrained(
                output_dir,
                state_dict=state_dict,
                safe_serialization=self.args.save_safetensors,
            )
        else:
            msg = f"Only support 'PretrainedRetriever' models but got: {self.model.__class__.__name__}"
            raise RuntimeError(msg)

        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))



"""From TRL Library:

https://github.com/huggingface/trl/blob/d57e4b726561e5ae58fdc335f34029052944a4a3/trl/trainer/model_config.py
"""

from dataclasses import dataclass, fields
from typing import Dict, List, Optional


@dataclass
class ModelArguments:
    """Arguments which define the model and tokenizer to load."""

    model_name_or_path: Optional[str] = None
    """The model checkpoint for weights initialization."""
    model_revision: Optional[str] = None
    """The specific model version to use (can be a branch name, tag name or commit id)."""
    torch_dtype: Optional[str] = None
    """Override the default `torch.dtype` and load the model under this dtype. If `auto` is passed, the
    dtype will be automatically derived from the model's weights.
    """
    trust_remote_code: bool = False
    """Trust remote code when loading a model."""
    attn_implementation: Optional[str] = None
    """Which attention implementation to use; you can run --attn_implementation=flash_attention_2, in which case you must install this manually by running `pip install flash-attn --no-build-isolation`"""
    encoder_class: Optional[str] = None
    """Name or alias of the PretrainedEncoder subclass that should be used as the encoder.
    If not specified, use the subclass of PretrainedEncoder that can load the given checkpoint.
    """
    pooling: Optional[str] = None
    """The type of pooling to use (options are 'mean', 'first_token', 'last_token'). Make sure the encoder class that you are using allows dynamically choosing the pooling method (i.e., supports setting this option)."""
    normalize: Optional[str] = None
    """Whether to normalize the embedding vector or not (options are 'yes' and 'no'). Make sure the encoder class that you are using allows selecting the normalization dynamically (i.e., supports setting this option)."""
    loss: Optional[str] = None
    """Name of the loss function to use for IR training."""
    temperature: float = 1.0
    """scaling factor for similarity scores when calculating the loss. We do not enforce its use.
    It is up to the loss function to use this argument.
    """
    temperature_learnable: bool = False
    """If true, make temperature a learnable parameter with initial value set to '--temperature' argument.
    WARNING: This feature is not complete yet and the learned temperature value is not saved to checkpoints. So you won't be able to load it later.
    """
    use_peft: bool = False
    """Whether to use PEFT or not for training."""
    lora_r: Optional[int] = 16
    """LoRA R value."""
    lora_alpha: Optional[int] = 32
    """LoRA alpha."""
    lora_dropout: Optional[float] = 0.05
    """LoRA dropout."""
    lora_target_modules: Optional[List[str]] = None
    """LoRA target modules."""
    lora_modules_to_save: Optional[List[str]] = None
    """Model layers to unfreeze & train"""
    lora_task_type: str = "FEATURE_EXTRACTION"
    """The task_type to pass for LoRA (use SEQ_CLS for reward modeling)"""
    use_rslora: bool = False
    """Use Rank-Stabilized LoRA (`paper <https://huggingface.co/papers/2312.03732>`_), which sets the adapter scaling factor to lora_alpha/√r, instead of the original default value of `lora_alpha/r`."""
    load_in_8bit: bool = False
    """use 8 bit precision for the base model - works only with LoRA"""
    load_in_4bit: bool = False
    """use 4 bit precision for the base model - works only with LoRA"""
    bnb_4bit_quant_type: Optional[str] = "nf4"
    """precise the quantization type (fp4 or nf4)"""
    use_bnb_nested_quant: bool = False
    """use nested quantization"""

    def __post_init__(self):
        if self.load_in_8bit and self.load_in_4bit:
            raise ValueError("You can't use 8 bit and 4 bit precision at the same time")

        if (
            isinstance(self.lora_target_modules, list)
            and len(self.lora_target_modules) == 1
        ):
            self.lora_target_modules = self.lora_target_modules[0]

        if self.use_rslora:
            raise NotImplementedError

        assert self.torch_dtype in ["auto", "bfloat16", "float16", "float32", None]
        assert self.normalize in ["yes", "no", None]

    def to_dict(self) -> Dict:
        """Return a json serializable view of the class attributes."""
        json_dict = dict()
        field_names = [f.name for f in fields(self)]
        for fname in field_names:
            fvalue = getattr(self, fname)
            json_dict[fname] = fvalue
        return json_dict



"""Base loss for information retrieval."""

from typing import Optional

import torch
from torch import nn

from .model_args import ModelArguments


class RetrievalLoss(nn.Module):
    """Base class for loss functions that can automatically detect the correct subclass to
    instantiate.

    To add a new loss function, create a new class that inherits from :class:`RetrievalLoss`.
    The users can instantiate the new loss function by its name.
    For example, if you do::

        class MyLoss(RetrievalLoss):
            _alias = 'foo_loss'
            ...

    Then users can use ``MyLoss`` if ``ModelArguments.loss == "MyLoss"`` or ``ModelArguments.loss == "foo_loss"``.

    **Do not overwrite or modify** ``cls._loss_registry`` class attribute.
    """

    # An alternative name (other than class name) that others can use to refer to your loss function.
    _alias: Optional[str] = None
    _loss_registry = dict()

    def __init_subclass__(cls, **kwargs):
        """Store a pointer to new loss functions that inherit from ``RetrievalLoss``."""
        super().__init_subclass__(**kwargs)
        assert cls.__name__ not in cls._loss_registry
        cls._loss_registry[cls.__name__] = cls
        if cls._alias is not None:
            assert cls._alias not in cls._loss_registry
            cls._loss_registry[cls._alias] = cls

    @classmethod
    def available_losses(cls) -> None:
        """Prints a list of all available loss functions and their aliases."""
        all_losses = list()
        for l in cls._loss_registry.values():
            if l not in all_losses:
                all_losses.append(l)
        if len(all_losses) == 0:
            print("Did not find any loss functions.")
            return
        print('Available losses in ("name", "alias") format:')
        for l in all_losses:
            print(f'("{l.__name__}", "{l._alias}")')

    @classmethod
    def from_model_args(cls, args: ModelArguments, **kwargs):
        """Instantiate the correct subclass of BaseLoss based on ``args.loss``.

        Args:
            args (ModelArguments): ``args.loss`` is used to detect the correct loss function subclass.
                ``args`` is also passed to the constructor of the target subclass.
            **kwargs: is passed to the constructor of the target subclass.

        Returns:
            An instance of the specified loss subclass. Or ``None`` if ``args.loss is None``
        """
        if args.loss is None:
            return None
        loss_cls = cls._loss_registry[args.loss]
        return loss_cls(args=args, **kwargs)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    def forward(
        self, logits: torch.Tensor, label: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Calculates the loss given the similarity scores between query and passages.

        The ``logits`` argument contains the similarity score between queries and
        passages (including in-batch negatives) for the entire batch.
        The shape of the ``logits`` argument is `[NUM_QUERIES, NUM_QUERIES * DOCS_PER_QUERY]`.
        The documents are organized sequentially. Basically, the related docs for the i_th query
        are ``docs[i * DOCS_PER_QUERY: (i + 1) * DOCS_PER_QUERY]`` and the rest are in-batch negatives.

        The ``label`` argument contains the groundtruth relevancy level between queries and documents for the entire
        batch. ``label`` does NOT include in-batch negatives. ``label`` tensor is of shape `[NUM_QUERIES, DOCS_PER_QUERY]`.
        We assign in-batch negatives, a relevancy level of 0 and extend the ``label`` argument accordingly.

        You can use the following snippet of code to make sure ``logits`` and ``labels`` are of the same shape and give
        a label of zero (0) to in-batch negatives.

        .. code:: python

            if label.size(1) != logits.size(1):
                label = torch.block_diag(*torch.chunk(label, label.shape[0]))

        Args:
            logits (torch.Tensor): The similarity scores between queries and passages (including in-batch negatives).
                shape: `[NUM_QUERIES, NUM_QUERIES * DOCS_PER_QUERY]`
            label (torch.Tensor): The groundtruth relevancy level between queries and passages (excluding in-batch negatives).
                shape: `[NUM_QUERIES, DOCS_PER_QUERY]`.
            **kwargs: Not used. Just to make the signature compatible with other losses.

        Returns:
            the loss value.
        """
        raise NotImplementedError



from typing import Dict, Optional

import torch
from transformers import TrainingArguments

from . import modeling_utils
from .model_args import ModelArguments
from .pretrained_encoder import PretrainedEncoder


class DefaultEncoder(PretrainedEncoder):
    _alias = "default"

    @classmethod
    def can_wrap(cls, *args, **kwargs) -> bool:
        return False

    def __init__(
        self,
        args: ModelArguments,
        training_args: Optional[TrainingArguments] = None,
        preprocess_only: bool = False,
        **kwargs,
    ) -> None:
        """A generic wrapper that can be used with any model.

        You can customize the pooling and normalization through ``model_args`` options.
        This wrapper does not do any custom formatting for queries and passages.
        It just returns the main text (prefix by the title if available).

        Args:
            args (ModelArguments): config for instantiating the model
            training_args (TrainingArguments): Not used by this wrapper.
            preprocess_only (bool): If True, do not load model parameteres and only
                provide methods and attributes necessary for pre-processing the input data.
            **kwargs: passed to ``transformers.AutoModel.from_pretrained``.
        """
        super().__init__(args, **kwargs)

        if args.normalize is None or args.pooling is None:
            msg = "You should specify both `normalize` and `pooling` options when using the default encoder wrapper."
            raise ValueError(msg)

        if args.normalize.lower() == "yes":
            self.normalize = True
        elif args.normalize.lower() == "no":
            self.normalize = False
        else:
            msg = f"'normalize' option must be one of `yes` or `no`. Got '{args.normalize}'"
            raise ValueError(msg)

        self.append_eos_token = False
        if args.pooling.lower() == "mean":
            self.pooling_fn = modeling_utils.mean_pooling
        elif args.pooling.lower() == "first_token":
            self.pooling_fn = modeling_utils.first_token_pooling
        elif args.pooling.lower() == "last_token":
            self.pooling_fn = modeling_utils.last_token_pooling
            self.append_eos_token = True
        else:
            msg = f"'pooling' option must be one of `mean`, `first_token`, or `last_token`. Got '{args.pooling}'"
            raise ValueError(msg)

        if not preprocess_only:
            self.model = modeling_utils.load_transformers_model(
                model_args=args, training_args=training_args, **kwargs
            )
            # The wrapped model should take care of these
            # Expose these to remain compatible with a subset (and *NOT* all) of tools
            # (e.g., transformers.Trainer) that expect a transformers.PretrainedModel instance
            modeling_utils.add_model_apis_to_wrapper(wrapper=self, model=self.model)
        else:
            self.model = None

    def encode(self, inputs: Optional[Dict[str, torch.Tensor]]) -> torch.Tensor:
        hidden_states = self.model(**inputs, return_dict=True)
        hidden_states = hidden_states.last_hidden_state
        att_mask = inputs["attention_mask"]

        embs = self.pooling_fn(last_hidden_state=hidden_states, attention_mask=att_mask)
        if self.normalize:
            embs = torch.nn.functional.normalize(embs, p=2, dim=-1)
        return embs

    def format_query(self, text: str, **kwargs) -> str:
        return text.strip()

    def format_passage(self, text: str, title: Optional[str] = None, **kwargs) -> str:
        if title is None:
            return text.strip()
        else:
            return f"{title.strip()} {text.strip()}".strip()

    def save_pretrained(self, *args, **kwargs):
        if "state_dict" in kwargs:
            prefix = "model."
            assert all(k.startswith(prefix) for k in kwargs["state_dict"].keys())
            kwargs["state_dict"] = {
                k[len(prefix) :]: v for k, v in kwargs["state_dict"].items()
            }
        return self.model.save_pretrained(*args, **kwargs)



"""Simple utilities to create python loggers."""

import logging
import os
from pathlib import Path
from typing import Optional, Tuple, Union

from rich.logging import RichHandler

from . import config

TROVE_LOGGING_FORMAT_STR_RICH = "%(funcName)s() %(message)s"
TROVE_LOGGING_FORMAT_STR = (
    "%(asctime)s %(module)s:%(lineno)s, [%(funcName)s] (%(levelname)s)- %(message)s"
)
TROVE_LOGGING_TIME_FORMAT_STR = "%H:%M:%S"

STR_LOG_LEVEL_TO_INT = {
    "CRITICAL": logging.CRITICAL,
    "DEBUG": logging.DEBUG,
    "ERROR": logging.ERROR,
    "FATAL": logging.FATAL,
    "INFO": logging.INFO,
    "NOTSET": logging.NOTSET,
    "WARN": logging.WARN,
    "WARNING": logging.WARNING,
}
INT_LOG_LEVEL_TO_STR = {v: k for k, v in STR_LOG_LEVEL_TO_INT.items()}


def ensure_str_log_level(level: Union[str, int]) -> str:
    """Ensures log level is valid and converts it to ``str`` levels recognized by python logging
    module."""
    if isinstance(level, str):
        if level.upper() not in STR_LOG_LEVEL_TO_INT:
            msg = (
                f"Log level can only be one of '{list(STR_LOG_LEVEL_TO_INT.keys())}'."
                f" Got '{level}'"
            )
            raise ValueError(msg)
        return level.upper()
    elif isinstance(level, int):
        if level not in STR_LOG_LEVEL_TO_INT.values():
            msg = (
                f"Log level can only be one of '{list(INT_LOG_LEVEL_TO_STR.keys())}'."
                f" Got '{level}'"
            )
            raise ValueError(msg)
        return INT_LOG_LEVEL_TO_STR[level]
    else:
        msg = f"Log level can only be 'int' or 'str'. Got '{type(level)}'"
        raise TypeError(msg)


class LoggerConfig:
    def __init__(self, level: Union[int, str]) -> None:
        """Hold information about a specific logger.

        Args:
            level (Union[int, str]): log level of the target logger.
        """
        self.level = ensure_str_log_level(level)

    def is_debug(self) -> bool:
        return STR_LOG_LEVEL_TO_INT[self.level] == logging.DEBUG

    def is_info(self) -> bool:
        return STR_LOG_LEVEL_TO_INT[self.level] == logging.INFO

    def is_warning(self) -> bool:
        return STR_LOG_LEVEL_TO_INT[self.level] == logging.WARNING


class TroveLoggerWrapper:
    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger

    def info(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.debug(*args, **kwargs)

    def warning(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.error(*args, **kwargs)

    def log(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.log(*args, **kwargs)

    def critical(self, *args, **kwargs):
        if not config.TROVE_LOGGING_DISABLE:
            return self._logger.critical(*args, **kwargs)


def rpath(path: Union[Path, str, os.PathLike]) -> str:
    """Make sure path starts wtih ``/`` or ``./`` so rich will highlight it in logs."""
    path = Path(path).as_posix()
    if not (path.startswith("/") or path.startswith(".")):
        path = "./" + path
    return path


def get_logger_with_config(
    name: Optional[str] = "trove",
    log_level: Optional[str] = None,
    rank: Optional[int] = None,
    force: bool = False,
) -> Tuple[Union[logging.Logger, TroveLoggerWrapper], LoggerConfig]:
    """Creates and returns a logger instance and a custom config object containing a subset of
    logging configurations.

    Args:
        name (Optional[str]): name of the logger to create.
        rank (Optional[int]): process rank in distributed environments
        log_level (Optional[str]): logging level. If None, it uses ``config.LOG_LEVEL``.
        force (bool): Only useful if log handler is already initialized. In that case
            If True, remove the current log handler and add a new one with the given arguments.
            If False, do nothing and return the current logger.

    Returns:
        a tuple of ``logger`` and ``TroveLoggingConfig``
    """
    logger = logging.getLogger(name)
    log_level = ensure_str_log_level(
        config.LOG_LEVEL if log_level is None else log_level
    )
    if force and len(logger.handlers):
        for h in logger.handlers[:]:
            logger.removeHandler(h)
    if not len(logger.handlers):
        # It is a new logger that needs to be initialized
        logger.setLevel(getattr(logging, log_level))
        if config.TROVE_LOGGING_DISABLE_RICH:
            handler = logging.StreamHandler()
            handler.setLevel(getattr(logging, log_level))
            format_str = TROVE_LOGGING_FORMAT_STR
            if rank is not None:
                format_str = format_str.replace("%(asctime)s", f"%(asctime)s <R{rank}>")
            if name != "trove":
                format_str = f"[{name}] " + format_str
            handler.setFormatter(
                logging.Formatter(format_str, TROVE_LOGGING_TIME_FORMAT_STR)
            )
        else:
            handler = RichHandler(
                level=getattr(logging, log_level),
                omit_repeated_times=False,
                locals_max_length=None,
                locals_max_string=None,
            )
            handler.setLevel(getattr(logging, log_level))
            format_str = TROVE_LOGGING_FORMAT_STR_RICH
            if rank is not None:
                format_str = f"<R{rank}> " + TROVE_LOGGING_FORMAT_STR_RICH
            if name != "trove":
                format_str = f"[{name}] " + format_str
            handler.setFormatter(
                logging.Formatter(format_str, TROVE_LOGGING_TIME_FORMAT_STR)
            )
        logger.addHandler(handler)
    logging_config = LoggerConfig(level=log_level)
    return logger, logging_config


# Create main Trove logger
_ = get_logger_with_config("trove")



from .materialized_qrel import MaterializedQRel
from .materialized_qrel_config import MaterializedQRelConfig
from .result_heapq import ResultHeapq
from .result_heapq_fast import FastResultHeapq
from .rows_by_key import RowsByKey, RowsByKeySingleSource



import copy
import itertools
import os
import random
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import datasets
from datasets.fingerprint import Hasher
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from transformers import Trainer

from .. import cache_manager, containers, file_utils
from ..containers import MaterializedQRel, MaterializedQRelConfig
from ..file_utils import JSONLinesWriter
from ..logging_utils import get_logger_with_config, rpath
from .data_args import DataArguments

logger, logging_conf = get_logger_with_config("trove")


class BinaryDataset(Dataset):
    def __init__(
        self,
        data_args: DataArguments,
        format_query: Callable[[str, Optional[str]], str],
        format_passage: Callable[[str, str, Optional[str]], str],
        positive_configs: Optional[
            Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]
        ] = None,
        negative_configs: Optional[
            Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]
        ] = None,
        train_cache_path: Optional[os.PathLike] = None,
        data_args_overrides: Optional[Dict[str, Any]] = None,
        trainer: Optional[Trainer] = None,
        outside_trainer: bool = False,
        num_proc: Optional[int] = None,
    ) -> None:
        """IR training dataset with only two levels of relevance (i.e., only positive and negative
        passages).

        * collections of positive (negative) passages are created from ``positive_configs`` (``negative_configs``) argument.
        * passages that appear in either of the positive or negative collections will be included in the resulting dataset.

        Args:
            data_args (DataArguments): general arguments for loading and processing the data
            format_query (Callable[[str, Optional[str]], str]): A callable that takes the query
                text and optionally the dataset name and returns the formatted query text for the modle.
            format_passage (Callable[[str, str, Optional[str]], str]): A callable that takes
                the passage text and title and dataset name and returns the formatted passage text for the model.
            positive_configs (Optional[Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]]):
                Config for one or multiple collections of queries, documents, and the relation between them.
                The passages from these collections are used as positives.
            negative_configs (Optional[Union[MaterializedQRelConfig, List[MaterializedQRelConfig]]]):
                Config for one or multiple collections of queries, documents, and the relation between them.
                The passages from these collections are used as negatives.
            train_cache_path (Optional[os.PathLike]): **DO NOT USE**. For internal operations only and not stable.
                If given, create a dataset only for training from this cache file.
                This is much more memory efficient compared to creating the dataset on-the-fly.
                You should use :meth:`export_and_load_train_cache` method to take advantage of this.
            data_args_overrides (Optional[Dict[str, Any]]): A mapping from a subset of ``DataArguments`` attribute names to
                their new values. These key values override the corresponding attributes in ``data_args`` argument.
                It is useful if you want to create multiple datasets from the same `DataArguments` instance
                but make small changes for each dataset without creating new ``DataArguments`` instances.
            trainer (Optional[Trainer]): An instance of the ``transformers.Trainer`` class.
                The random seed and epoch from trainer instance are used to sample positive and negative documents.
            outside_trainer (bool): If true, do not use trainer instance and set seed and epoch both to zero.
                Useful for debugging without a trainer instance.
            num_proc (Optional[int]): arg to to methods like ``datasets.Dataset.*``
        """
        super().__init__()
        # A unique fingerprint for the underlying data for this dataset.
        # Note that two datasets with the same fingerprint are only based on the same data
        # but can generate different samples. see docstring for self.fingerprint for details.
        # do not use this directly. Instead use 'self.fingerprint' attribute
        self._unique_fingerprint = None
        # some metadata about this dataset to be saved along with any cache files if necessary
        # Do not use this directly. Instead, use 'self.info' attribute
        self._info = None
        # If true, MultiLevelDataset[i] returns the training group (i.e., output of
        # self.create_group() which just returns the related docs) for each query rather than a training sample.
        # this is useful for iterating over the underlying data for each dataset instance.
        self._iter_over_groups = False

        if (positive_configs is None) != (negative_configs is None):
            msg = (
                "You should pass either both or none of the 'positive_configs' and 'negative_configs' arguments."
                f" Got: 'positive_configs': '{positive_configs}' and 'negative_configs': '{negative_configs}'"
            )
            raise ValueError(msg)

        if (positive_configs is not None or negative_configs is not None) == (
            train_cache_path is not None
        ):
            msg = (
                "You should pass exactly one of the ('positive_configs', 'negative_configs') pair or 'train_cache_path' arguments."
                f" Got: 'positive_configs': '{positive_configs}', 'negative_configs': '{negative_configs}', and 'train_cache_path': '{train_cache_path}'"
            )
            raise ValueError(msg)

        # make a deepcopy so we do not impact operations outside of this class.
        data_args = copy.deepcopy(data_args)
        if data_args_overrides is not None:
            for attr, val in data_args_overrides.items():
                if not hasattr(data_args, attr):
                    # We should not add new attributes
                    msg = f"'{attr}' is not a valid attribute name for 'DataArguments'"
                    raise ValueError(msg)
                setattr(data_args, attr, val)
        self.args = data_args

        self.format_query = format_query
        self.format_passage = format_passage

        self.outside_trainer = outside_trainer
        self.trainer = trainer

        # by default, the training groups should be created on-the-fly
        self.create_group = self.create_group_on_the_fly

        self.train_records = None
        if train_cache_path is not None:
            logger.info(
                f"Creating BinaryDataset from cached training groups in {rpath(train_cache_path)}."
            )
            if Path(train_cache_path).suffix != ".jsonl":
                msg = f"Only supported format for 'train_cache_path' is json lines. Got: {train_cache_path}"
                raise ValueError(msg)
            self.train_records = datasets.load_dataset(
                "json", data_files=Path(train_cache_path).as_posix(), split="train"
            )
            self.create_group = self.create_group_from_cache
            return

        # Make sure everything is of type 'list'
        if not isinstance(positive_configs, list):
            positive_configs = [positive_configs]
        if not isinstance(negative_configs, list):
            negative_configs = [negative_configs]

        # This is for logging purposes only. Do NOT use
        self._positive_configs = positive_configs
        self._negative_configs = negative_configs

        _all_files = list()
        for conf in positive_configs + negative_configs:
            conf.ensure_list_of_correct_dtype()
            if conf.query_path is not None:
                if isinstance(conf.query_path, list):
                    _all_files.extend(conf.query_path)
                else:
                    _all_files.append(conf.query_path)

            if conf.corpus_path is not None:
                if isinstance(conf.corpus_path, list):
                    _all_files.extend(conf.corpus_path)
                else:
                    _all_files.append(conf.corpus_path)

        # Create a unique file id for each file path used in this dataset
        # This is used by MaterializedQRel instances to create globally unique query/passage IDs across files
        # which allows us to merge various files without concerns about duplicate IDs
        fname_to_id_map = file_utils.create_file_name_to_id_mapping(
            files=_all_files, id_length="shortest"
        )

        all_record_mappings = dict()
        for path in _all_files:
            _realpath = Path(path).resolve().absolute().resolve().as_posix()
            if _realpath not in all_record_mappings:
                all_record_mappings[_realpath] = containers.RowsByKeySingleSource(
                    filepath=path,
                    new_key_field="_id",
                    add_filepath_field=True,
                    num_proc=num_proc,
                )

        # Create the necessary qrel collections for both positive and negative documents.
        logger.info("Load MaterializedQRel for positive passages.")
        self.positive_passages = [
            MaterializedQRel(
                args=item,
                file_name_to_id=fname_to_id_map,
                record_mapping_collection=all_record_mappings,
                num_proc=num_proc,
            )
            for item in positive_configs
        ]
        logger.info("Load MaterializedQRel for negative passages.")
        self.negative_passages = [
            MaterializedQRel(
                args=item,
                file_name_to_id=fname_to_id_map,
                record_mapping_collection=all_record_mappings,
                num_proc=num_proc,
            )
            for item in negative_configs
        ]

        qids_with_pos = list(
            itertools.chain.from_iterable(
                [p.all_qids_global for p in self.positive_passages]
            )
        )
        qids_with_neg = list(
            itertools.chain.from_iterable(
                [p.all_qids_global for p in self.negative_passages]
            )
        )

        # Queries that have both positive and negative passages and thus can be used by this dataset
        # Sort them for consistency across different runs
        self.all_qids = list(
            sorted(list(set(qids_with_pos).intersection(set(qids_with_neg))))
        )

    def update_metadata(self) -> None:
        """Updates the metadata for the dataset.

        It creates a new fingerprint and metadata dict for the dataset.
        """
        # The underlying data for each dataset is the set of MaterializedQRel collections
        # So its fingerprint is also the combined fingerprint of the MaterializedQRel collections that it contains.
        hasher = Hasher()
        for qc in self.negative_passages:
            hasher.update(qc.fingerprint)
        for qc in self.positive_passages:
            hasher.update(qc.fingerprint)
        self._unique_fingerprint = hasher.hexdigest()
        self._info = {
            "fingerprint": self._unique_fingerprint,
            "data_args": self.args.to_dict(),
            "positive_configs": [qc.to_dict() for qc in self._positive_configs],
            "negative_configs": [qc.to_dict() for qc in self._negative_configs],
        }

    @property
    def fingerprint(self) -> str:
        """Calculates a unique fingerprint for the contents and output of this dataset.

        Datasets with the same fingerprint are backed by the same underlying data but do **NOT**
        necessarily generate the same samples. For example, using different sampling strategies
        or query and document formatting functions leads to different output from the same underlying
        data and thus the same fingerprint.

        This fingerprint is for internal operations only and you should not rely on it. And if you do,
        just use it to identify the underlying data (e.g., for caching and loading) and not
        the exact samples.

        datasets with the same fingerprint are backed by the same data sources.
        """
        if self._unique_fingerprint is None:
            self.update_metadata()
        return self._unique_fingerprint

    @property
    def info(self) -> Dict:
        if self._info is None or self._info["fingerprint"] != self.fingerprint:
            self.update_metadata()
        return self._info

    def set_index_lookup_storage_type(self, storage: str) -> None:
        """Select if the key to row index lookup table should be stored in memory or in memory-
        mapped lmdb dict."""
        for qc in self.negative_passages:
            qc.set_index_lookup_storage_type(storage)
        for qc in self.positive_passages:
            qc.set_index_lookup_storage_type(storage)

    def create_group_from_cache(self, index: int) -> Dict[str, Union[str, List[Dict]]]:
        """Loads one query and the related negative and positive passages.

        Input and output is the same as :meth:`create_group_on_the_fly` but reads the data from
        cache file rather than creating it on-the-fly.
        """
        return self.train_records[index]

    def create_group_on_the_fly(self, index: int) -> Dict[str, Union[str, List[Dict]]]:
        """Loads one query and the related negative and positive passages.

        The content of query/passages is also loaded (i.e., records are materialized.).
        The return format is based on `tevatron <https://github.com/texttron/tevatron>`_ .

        Args:
            index (int): Index of the query to load

        Returns:

            A dict of the following format::

                {
                    'query_id': 'Unique ID of query across all files in this dataset'.
                    'query': 'query text',
                    'positive_passages': [ # list of positive documents for this query
                    {'_id': 'globally unique id of the passage', 'text': '...', 'title': '...'}, # There could be additional fields in this dict, which should be ignored
                    ...,
                    {'_id': ....}
                    ],
                    'negative_passages': [ # list of negative documents for this query
                    {'_id': ....}, # The same data structure and field names as positive documents.
                    ...,
                    {'_id': ....}
                    ]

                }
        """
        qid = self.all_qids[index]

        # Collect query/passage records in dicts from `_id` to `record` to avoid duplicate records
        q_recs = dict()
        pos_recs = dict()
        neg_recs = dict()

        # Load negative passages from different sources
        for neg_col in self.negative_passages:
            query_rec, doc_recs = neg_col.get_related_recs(
                global_qid=qid, materialize=True, return_global_ids=True, strict=False
            )
            if query_rec is None or doc_recs is None:
                # this collection does not have entries for this specific query
                continue
            q_recs[query_rec["_id"]] = query_rec
            for d_rec in doc_recs:
                neg_recs[d_rec["_id"]] = d_rec

        # Load positive passages from different sources
        for pos_col in self.positive_passages:
            query_rec, doc_recs = pos_col.get_related_recs(
                global_qid=qid, materialize=True, return_global_ids=True, strict=False
            )
            if query_rec is None or doc_recs is None:
                # this collection does not have entries for this specific query
                continue
            q_recs[query_rec["_id"]] = query_rec
            for d_rec in doc_recs:
                pos_recs[d_rec["_id"]] = d_rec

        # Make sure all collections return the same data for this query
        assert len(q_recs) == 1
        q_rec = next(iter(q_recs.values()))

        if (
            len(set(list(pos_recs.keys())).intersection(set(list(neg_recs.keys()))))
            != 0
        ):
            msg = "Some documents are used as both positive and negatives for the same query."
            raise RuntimeError(msg)

        pos_docs = list(pos_recs.values())
        neg_docs = list(neg_recs.values())

        group = {
            "query_id": q_rec["_id"],
            "query": q_rec["text"],
            "positive_passages": pos_docs,
            "negative_passages": neg_docs,
        }

        return group

    def __len__(self) -> int:
        if self.train_records is not None:
            return len(self.train_records)
        else:
            return len(self.all_qids)

    def set_trainer(self, trainer: Trainer) -> None:
        """Set the trainer attribute."""
        if self.trainer is None:
            self.trainer = trainer

    def epoch_and_seed(self) -> Tuple[int, Union[float, int]]:
        """If trainer instance is available, load seed and current epoch from trainer."""
        if self.outside_trainer:
            epoch = 0
            seed = 0
        else:
            epoch = int(self.trainer.state.epoch)
            seed = self.trainer.args.seed
        return epoch, seed

    def __getitem__(
        self, index: int
    ) -> Dict[str, Union[str, List[str], List[Union[int, float]]]]:
        """Create a training example consisting of one query, one positive passage and multiple
        negative passages.

        taken from `here <https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/dataset.py#L41>`_
        With minor changes to make it work with this class.

        Returns:

            a dict with the following format::

                {
                    'query': formatted query of type `str`,
                    'passage': a list of formatted passages for the query.
                        The positive passage is the first element of the list and the rest are negatives.
                    'labels': the relevancy level of each passage (i.e., score).
                        Positive and negative passages are given scores of 1 and 0, respectively.
                }
        """
        group = self.create_group(index)
        if self._iter_over_groups:
            # we just need to return the training groups (and not training instances) without any processing
            return group
        epoch, trainer_seed = self.epoch_and_seed()

        _hashed_seed = hash(index + trainer_seed)

        query = group["query"]
        group_positives = group["positive_passages"]
        group_negatives = group["negative_passages"]

        formatted_query = self.format_query(text=query, dataset=self.args.dataset_name)
        formatted_passages = []
        passage_scores = list()

        if self.args.positive_passage_no_shuffle:
            pos_psg = group_positives[0]
        else:
            pos_psg = group_positives[(_hashed_seed + epoch) % len(group_positives)]

        formatted_passages.append(
            self.format_passage(
                text=pos_psg["text"],
                title=pos_psg["title"],
                dataset=self.args.dataset_name,
            )
        )
        passage_scores.append(1)

        negative_size = self.args.group_size - 1
        if len(group_negatives) < negative_size:
            negs = random.choices(group_negatives, k=negative_size)
        elif self.args.group_size == 1:
            negs = []
        elif self.args.negative_passage_no_shuffle:
            negs = group_negatives[:negative_size]
        else:
            _offset = epoch * negative_size % len(group_negatives)
            negs = [x for x in group_negatives]
            random.Random(_hashed_seed).shuffle(negs)
            negs = negs * 2
            negs = negs[_offset : _offset + negative_size]

        for neg_psg in negs:
            formatted_passages.append(
                self.format_passage(
                    text=neg_psg["text"],
                    title=neg_psg["title"],
                    dataset=self.args.dataset_name,
                )
            )
            passage_scores.append(0)

        output = {
            "query": formatted_query,
            "passage": formatted_passages,
            "label": passage_scores,
        }
        return output

    def export_and_load_train_cache(
        self,
        cache_file: Optional[os.PathLike] = None,
        cache_pardir: Optional[os.PathLike] = None,
        num_proc: Optional[int] = None,
        batch_size: int = 16,
    ):
        """Export the training groups to a cache file and load them into a new instance of
        BinaryDataset.

        To reduce memory consumption, it generates all the training groups, write them
        into a json lines file and returns a new ``BinaryDataset`` instance from those cached records.
        To benefit from the reduced memory consumption, make sure you do not keep any references to the old dataset instance,
        so it can be garbage collected by the interpreter.

        You can do something like::

            dataset = BinaryDataset(...)
            dataset = dataset.export_and_load_train_cache()
            gc.collect() # if you want to force the interpreter to release the memory right away

        Args:
            cache_file: a json lines files to save the cached training groups. If ``None``, a unique cache
                file is created based on the dataset fingerprint.
            cache_pardir: the parent directory to save the cache file to. If provided, we create a
                a subdir in this directory based on the dataset fingerprint and save the dataset cache
                in that subdir.
            num_proc: number of workers to use to generate the training groups.
            batch_size: read the training groups in batches of the given size

        Returns:
            A new instance of ``BinaryDataset`` for training that is backed by the cached training groups.
        """
        logger.info("Export BinaryDataset training groups to cache file.")
        if num_proc is None:
            num_proc = 0

        if cache_file is not None and cache_pardir is not None:
            msg = (
                "You can pass at most one of the 'cache_file' or 'cache_pardir' arguments."
                f" Got: 'cache_pardir': '{cache_pardir}' and 'cache_file': '{cache_file}'"
            )
            raise ValueError(msg)

        if cache_file is None:
            if cache_pardir is not None:
                cache_file = cache_manager.get_cache_dir(
                    cache_pardir=cache_pardir,
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
            else:
                cache_file = cache_manager.get_cache_dir(
                    artifact_content="binary_dataset_train_cache",
                    artifact_type="final",
                    fingerprint=self.fingerprint,
                    metadata=self.info,
                )
            cache_file = Path(cache_file, "train_records.jsonl")
        else:
            cache_file = Path(cache_file)
        cache_file.parent.mkdir(exist_ok=True, parents=True)

        with file_utils.easyfilelock(cache_file.as_posix()):
            if not cache_file.exists():
                # We don't want to save the exact output of the dataset since that depends on sampling strategy, etc.
                # Those operations must be done during training and do not have a noticeable impact memory consumption
                orig_iter_groups = self._iter_over_groups
                self._iter_over_groups = True

                # use torch data loaders parallelize and speedup the read operations
                data_loader = DataLoader(
                    self,
                    batch_size=batch_size,
                    collate_fn=lambda x: x,
                    num_workers=num_proc,
                    drop_last=False,
                    shuffle=False,
                )
                with file_utils.atomic_write(file=cache_file, root="parent") as tfile:
                    with JSONLinesWriter(tfile) as rec_writer:
                        for batch in tqdm(
                            data_loader, desc="Export train groups to jsonl file"
                        ):
                            rec_writer.add(batch)
                # revert the dataset to its original format/functionality
                self._iter_over_groups = orig_iter_groups
        # do not forget to deepcopy attributes that are part of this instance (and not references to other objects)
        # If you do not do so, this instance will not get garbage collected, which makes the entire caching pointless
        return BinaryDataset(
            data_args=copy.deepcopy(self.args),
            format_query=self.format_query,
            format_passage=self.format_passage,
            train_cache_path=cache_file,
            trainer=self.trainer,
            outside_trainer=self.outside_trainer,
            num_proc=num_proc,
        )



"""From here:

https://github.com/huggingface/datasets/blob/a1b5a32365cd2c6631fa4ad236a881f9f62bf32b/src/datasets/commands/__init__.py
"""

from abc import ABC, abstractmethod
from argparse import ArgumentParser


class BaseTroveCLICommand(ABC):
    @staticmethod
    @abstractmethod
    def register_subcommand(parser: ArgumentParser):
        raise NotImplementedError()

    @abstractmethod
    def run(self):
        raise NotImplementedError()



"""This file contains base wrapper for encoder architectures.

See 'PretrainedEncoder' docstring for the primary goal of these wrapper classes and how to add a
new one.
"""

from typing import Dict, Optional

import torch
from torch import nn
from transformers import TrainingArguments

from . import modeling_utils
from .model_args import ModelArguments


class PretrainedEncoder(nn.Module):
    """A wrapper around different encoders.

    This class wraps the encoder and takes care of model specific actions like
    saving and loading checkpoints, pooling, normalization, formatting inputs, etc.

    ``PretrainedEncoder`` automatically detects and instantiates the correct wrapper subclass that can load a specific model.

    To support a new type of encoder, inherit from this class and implement the class method
    ``cls.can_wrap()``. This method takes the model name and its arguments and
    if it can wrap this model, returns True. Otherwise, returns False.

    **Do not overwrite or modify** ``cls._model_registry``.
    """

    _model_registry = list()
    _alias: Optional[str] = None

    def __init_subclass__(cls, **kwargs):
        """Store a pointer to new wrappers that inherit from `PretrainedEncoder`."""
        super().__init_subclass__(**kwargs)
        # prioritize user-defined wrappers over the ones defined by the library.
        if "trove" in cls.__module__.split("."):
            cls._model_registry.append(cls)
        else:
            cls._model_registry.insert(0, cls)

    @classmethod
    def can_wrap(cls, model_name_or_path: str, args: ModelArguments) -> bool:
        """returns true if this wrapper can wrap the specified model with the given arguments.

        Subclasses of ``PretrainedEncoder`` should implement this method.
        We use this method to automatically choose the correct subclass to wrap different models.

        Args:
            model_name_or_path (str): name of the model to wrap.
            args (ModelArguments): arguments that describe the model to wrap.

        Returns:
            True if this class can wrap the model, and false otherwise.
        """
        raise NotImplementedError

    @classmethod
    def find_appropriate_wrapper(cls, model_name_or_path: str, args: ModelArguments):
        """Find the appropriate wrapper than can wrap and load a specific model.

        If ``args.encoder_class`` is set, then a subclass of ``PretrainedEncoder`` with that name (or alias) is returned.
        Otherwise, the model arguments are passed to the ``can_wrap()`` method of all the registered subclasses of ``PretrainedEncoder``.
        The subclass that its ``can_wrap()`` method returns ``True`` is returned by this method.
        If the ``can_wrap()`` method of all subclasses return ``False`` and the checkpoint is a fine-tuned model, it finds the base model of the given checkpoint and
        repeats the same process to find the subclass that can load the base model.

        Args:
            model_name_or_path (str): name or path of the model that we want to wrap.
            args (ModelArguments): arguments that describe the model that we want to wrap.
                ``cls.can_wrap()`` method of subclasses might use this in addition
                to ``model_name_or_path`` to determine if they can wrap the model.

        Returns:
            A pointer to the subclass that can wrap the given model.
        """
        if args.encoder_class is not None and args.encoder_class.lower() != "none":
            # find the encoder wrapper with the given name
            encoder_cls = None
            for model_cls in cls._model_registry:
                if (
                    model_cls.__name__ == args.encoder_class
                    or model_cls._alias == args.encoder_class
                ):
                    encoder_cls = model_cls
                    break
            if encoder_cls is None:
                msg = f"Could not find a PretrainedEncoder subclass with the name or alias equal to '{args.encoder_class}'"
                raise ValueError(msg)
            return encoder_cls

        assert model_name_or_path is not None

        # List of classes that can load the model
        compatible_classes = list()
        for model_cls in cls._model_registry:
            if model_cls.can_wrap(model_name_or_path=model_name_or_path, args=args):
                compatible_classes.append(model_cls)
        if len(compatible_classes) > 1:
            msg = (
                "More than one wrapper can load this model. We cannot decide which one to use."
                f" Compatible classes: {compatible_classes}"
            )
            raise RuntimeError(msg)

        if len(compatible_classes) == 1:
            return compatible_classes[0]

        # None of the wrappers can load this model
        # Try to use the wrapper that can load the base model of the specified model.
        base_model_name = modeling_utils.find_base_name_or_path(model_name_or_path)
        if base_model_name is None:
            msg = "None of the encoder wrappers can load this model or any of its base models."
            raise RuntimeError(msg)

        encoder_cls = cls.find_appropriate_wrapper(
            model_name_or_path=base_model_name, args=args
        )
        return encoder_cls

    @classmethod
    def from_model_args(
        cls,
        args: ModelArguments,
        model_name_or_path: Optional[str] = None,
        training_args: Optional[TrainingArguments] = None,
        **kwargs,
    ):
        """Instantiate the correct subclass of ``PretrainedEncoder`` that can wrap the specified
        model.

        Args:
            args (ModelArguments): arguments that describe the model that we want to wrap.
            model_name_or_path (Optional[str]): name of the model to wrap.
                If not provided, use ``args.model_name_or_path``.
            training_args (TrainingArguments): passed to `__init__`. Used for activating gradient checkpointing.
            **kwargs: extra keyword arguments passed to the wrapper constructor.

        Returns:
            an instance of the correct wrapper class that wraps the specified model.
        """
        if model_name_or_path is None:
            model_name_or_path = args.model_name_or_path
        encoder_cls = cls.find_appropriate_wrapper(
            model_name_or_path=model_name_or_path, args=args
        )
        return encoder_cls(args=args, training_args=training_args, **kwargs)

    def __init__(
        self,
        args: ModelArguments,
        training_args: Optional[TrainingArguments] = None,
        preprocess_only: bool = False,
        **kwargs,
    ) -> None:
        """Wraps encoder models and provides methods and attributes for pre-processing encoder
        inputs.

        If ``preprocess_only`` is True, you are expected to only provide attributes and methods
        required for preparing the input for the encoder (e.g., ``append_eos_token``, ``format_query()``, etc.).
        If possible you should avoid loading the model parameters if ``preprocess_only`` is True.
        This allows us to pre-process the data without loading the model, which leaves
        more resources (e.g., memory) for preprocessing operations.

        Args:
            args (ModelArguments): config for instantiating the model
            preprocess_only (bool): if true, do not load the model parameters
                and just provide the attributes and methods necessary for pre-processing the input.
                E.g., ``append_eos_token``, ``format_query()``, etc.
            training_args (TrainingArguments): You can use this to activate gradient checkpointing if needed.
            **kwargs: extra kwargs passed to PretrainedModel.from_pretrained when loading
                the encoder model.
        """
        super().__init__()
        self.args = args
        self.config = None  # You should set this to config of the encoder (PretrainedModel.config)

    def encode(self, inputs: Optional[Dict[str, torch.Tensor]]) -> torch.Tensor:
        """calculate the embeddings for tokenized input."""
        raise NotImplementedError

    def format_query(self, text: str, **kwargs) -> str:
        """Format the query before passing it to tokenizer.

        You can also ask for other parameters like dataset_name for example if your model uses
        different formatting for different datasets like `intfloat/e5-mistral-7b-instruct`
        """
        raise NotImplementedError

    def format_passage(
        self,
        text: str,
        title: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Format the passage before passing it to tokenizer.

        You can also ask for other parameters like dataset_name for example if your model uses
        different formatting for different datasets like `intfloat/e5-mistral-7b-instruct`
        """
        raise NotImplementedError

    def save_pretrained(self, *args, **kwargs):
        """Save model parameters.

        It should replicate the signature and behavior of
        ``transformers.PreTrainedModel.save_pretrained``.
        """
        raise NotImplementedError

    def encode_query(self, inputs: Optional[Dict[str, torch.Tensor]]) -> torch.Tensor:
        """Overwrite if queries are encoded differently."""
        return self.encode(inputs=inputs)

    def encode_passage(self, inputs: Optional[Dict[str, torch.Tensor]]) -> torch.Tensor:
        """Overwrite if passages are encoded differently."""
        return self.encode(inputs=inputs)

    def similarity_fn(self, query: torch.Tensor, passage: torch.Tensor) -> torch.Tensor:
        """Similarity between query and passage embeddings.

        Overwrite if your encoder uses a different similarity function.

        Args:
            query (torch.Tensor): query embeddings. shape: `[NUM_QUERIES, EMB_DIM]`
            passage (torch.Tensor): passage embeddings. shape: `[NUM_PASSAGES, EMB_DIM]`

        Returns:
            query-passage similarities. shape is `[NUM_QUERIES, NUM_PASSAGES]`
        """
        return query @ passage.T

    def compute_scores(
        self, query: Dict[str, torch.Tensor], passage: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Compute similarity score between tokenized query and passages.

        Args:
            query (Dict[str, torch.Tensor]): query tokens
            passage (Dict[str, torch.Tensor]): passage tokens

        Returns:
            query-passage similarities. shape is `[NUM_QUERIES, NUM_PASSAGES]`
        """
        query_feats = self.encode_query(inputs=query)
        passage_feats = self.encode_passage(inputs=passage)
        scores = self.similarity_fn(queries=query_feats, passage=passage_feats)
        return scores

    def gradient_checkpointing_enable(self, *args, **kwargs):
        return self.model.gradient_checkpointing_enable(*args, **kwargs)



"""Manage global library configuration."""

import os
from contextlib import contextmanager
from uuid import uuid4

from datasets.fingerprint import Hasher

## Defaults

LOG_LEVEL = "WARNING"
# disable rich log handler
TROVE_LOGGING_DISABLE_RICH = False
# disable logging module
TROVE_LOGGING_DISABLE = False

# incomplete files are written to this directory (from file_utils.atomic_write function)
TROVE_INCOMPLETE_FILE_DIRNAME = "trove_incomplete_files"

# root directory where all cache files are saved.
# we pass this value to 'hf_hub.cached_assets_path(assets_dir=config.TROVE_CACHE_DIR)'
# If you set this to None, the cache dir is determined by hf_hub library
# (current it gives us a subdir in 'HUB_CACHE/assets')
TROVE_CACHE_DIR = None

# Create a unique ID for this specific process
# You can use it for example to create a temporary directory for
# all temporary files in this process
hasher = Hasher()
hasher.update(os.getpid())
hasher.update(uuid4().hex)
PROCESS_UUID = hasher.hexdigest()
del hasher

## Update default values if necessary

if os.getenv("TROVE_LOG_LEVEL", None) is not None:
    LOG_LEVEL = os.environ["TROVE_LOG_LEVEL"]

if os.getenv("TROVE_LOGGING_DISABLE_RICH", None) is not None:
    TROVE_LOGGING_DISABLE_RICH = (
        os.environ["TROVE_LOGGING_DISABLE_RICH"].lower() == "true"
    )

if (
    os.getenv("TROVE_CACHE_DIR", None) is not None
    and os.environ["TROVE_CACHE_DIR"].strip() != ""
):
    TROVE_CACHE_DIR = os.environ["TROVE_CACHE_DIR"]


def disable_logging() -> None:
    global TROVE_LOGGING_DISABLE
    TROVE_LOGGING_DISABLE = True


def enable_logging() -> None:
    global TROVE_LOGGING_DISABLE
    TROVE_LOGGING_DISABLE = False


@contextmanager
def logging_context(disable: bool = False):
    global TROVE_LOGGING_DISABLE
    orig_value = TROVE_LOGGING_DISABLE
    try:
        TROVE_LOGGING_DISABLE = disable
        yield
    finally:
        TROVE_LOGGING_DISABLE = orig_value



from dataclasses import dataclass
from typing import Optional

import torch.distributed as dist
from transformers import TrainingArguments

from ..logging_utils import get_logger_with_config


@dataclass
class EvaluationArguments(TrainingArguments):
    output_dir: Optional[str] = None
    """Directory to save the results."""
    per_device_matmul_batch_size: int = 256
    """Batch size for score calculation operation (i.e., ``matmul(q, doc)``).
    We use ``per_device_eval_batch_size`` as batch size for encoding the query and documents.
    """
    precompute_corpus_embs: Optional[bool] = False
    """Precompute the corpus embeddings, write to cache, and then calculate their score with query embeddings.
    They will be written to cache anyways (either temporarily or permanently).
    This option just controls if we should finish encoding documents before calculating the scores.
    """
    encoding_cache_dir: Optional[str] = None
    """If provided, write the embedding vectors to this directory.
    This option will be ignored for any ``EncodingDataset`` that has a cache filename attached to it already.
    If not provided, cache is written to a temporary directory and deleted before exit.
    """
    ir_metrics_k_values: Optional[str] = "1,3,5,10,100"
    """A comma separated list of cutoff values for IR metrics.
    **NOTE**: It is only used if ``compute_metrics`` is not passed to ``RetrievalEvaluator.__init__()`` method.
    """
    ir_metrics_relevance_threshold: Optional[int] = None
    """Minimum groundtruth relevancy level (inclusive) for a document to be considered relevant when calculating IR metrics.
    If not ``None``, it is passed to :class:`~trove.evaluation.metrics.IRMetrics` init method. See its docstring for details.
    **NOTE**: It is only used if ``compute_metrics`` is not passed to ``RetrievalEvaluator.__init__`` method.
    """
    search_topk: Optional[int] = None
    """Number of documents to retrieve during nearest neighbor search. Must be ``>= max(ir_metrics_k_values)``.
    Defaults to ``max(ir_metrics_k_values)``. This is useful to select the number of mined hard negatives for each query."""
    no_annot_in_mined_hn: bool = True
    """If true, annotated documents are **not** included in hard negative mining results, even if they are annotated as irrelevant.
    I.e., documents with groundtruth relevance label of zero are also excluded from hard negative mining results. If false, all documents are used for hard negative mining."""
    merge_mined_qrels: bool = False
    """By default, hard negative mining results are saved in a separate file for each pair of input query-corpus files.
    For example if you have two query files and two corpus files, you will end up with four qrel files.
    This allows you to combine multiple query (corpus) files that use the same ID different queries (documents).
    Set this to ``True`` to write all hard negative mining results in one file. It will raise an exception if it finds two records with the same ID.
    """
    pbar_mode: Optional[str] = "all"
    """Determines which processes show the progress bar. You can choose from one of ``['none', 'main', 'local_main', and 'all']`` values."""
    print_mode: Optional[str] = "main"
    """Determines which processes can print to stdout. You can choose from one of ``['none', 'main', 'local_main', and 'all']`` values."""
    cleanup_temp_artifacts: bool = True
    """If true, it removes all embedding cache files that it has generated but the user did not ask to save them explicitly."""
    save_eval_topk_logits: bool = False
    """If true, save the score of topk retrieved docs for each query during evaluation to disk. Note that if you have multiple query (or document) files that share the same ID, setting this to ``True`` will raise an exception."""
    output_qrel_format: str = "tsv"
    """The format and structure of the output file that search results (qid, docid, and scores) will be written to.
    There are two options:

        * ``tsv`` is the standard qrel format in a tsv file with three columns: `query-id`, `corpus-id`, and `score`
        * ``grouped`` is a json lines file where each record has three keys. ``qid`` is the ID of the current query.
          ``docid`` and ``score`` are two lists of the same size that hold the ID and score of the related documents for ``qid``
    """
    fair_sharding: bool = False
    """(Only used in distributed environments) If false, shard the dataset into chunks of roughly equal sizes.
    If true, shard the dataset such that devices with higher throughput are assigned bigger shards.
    This is to avoid idle GPU cycles when mixing GPUs with different capabilities.
    """
    broadcast_output: bool = True
    """(only for distributed environments) If true, the output of ``RetrievalEvaluator.evaluate()``
    and ``RetrievalEvaluator.mine_hard_negatives()`` are duplicated across all processes
    (i.e., these methods return identical outputs in all processes). If false, only
    the main process returns the output and other processes return ``None``. Set it to ``False`` to
    save memory on machines with multiple GPUs.
    """
    trove_logging_mode: str = "all"
    """Determines which processes can use the logging module. It is just a soft limit:
    the excluded processes can still log messages but their logging level is set to ERROR.
    You can choose from one of ``['main', 'local_main', and 'all']`` values.
    """

    def __post_init__(self):
        super().__post_init__()

        if self.trove_logging_mode not in ["main", "local_main", "all"]:
            msg = "Valid values for mode configs (print and pbar) are ['main', 'local_main', and 'all']"
            raise ValueError(msg)

        if dist.is_initialized():
            world_size = dist.get_world_size()
        else:
            world_size = 1
        if world_size < 2:
            return
        log_level = None
        if (self.trove_logging_mode == "main" and self.process_index != 0) or (
            self.trove_logging_mode == "local_main" and self.local_process_index != 0
        ):
            log_level = "ERROR"
        get_logger_with_config(
            name="trove", log_level=log_level, rank=self.process_index, force=True
        )



from typing import Dict, Optional

import torch
from transformers import AutoModel, PreTrainedModel, TrainingArguments

from . import modeling_utils
from .model_args import ModelArguments
from .pretrained_encoder import PretrainedEncoder


class ContrieverEncoder(PretrainedEncoder):
    @classmethod
    def can_wrap(cls, model_name_or_path: str, args: ModelArguments) -> bool:
        """Returns true if the model is an instance of ``facebook/contriever`` or
        ``facebook/contriever-msmarco`` models.

        We can also wrap models that their base model is one of these. But, :class:`~trove.modeling.pretrained_encoder.PretrainedEncoder` takes
        care of that. We just handle the main models.
        """
        _can_wrap = model_name_or_path in [
            "facebook/contriever",
            "facebook/contriever-msmarco",
        ]
        return _can_wrap

    def __init__(
        self,
        args: ModelArguments,
        training_args: Optional[TrainingArguments] = None,
        preprocess_only: bool = False,
        **kwargs,
    ) -> None:
        """Wraps contriever variants and also provides necessary attributes and methods for data
        pre-processing.

        Args:
            args (ModelArguments): config for instantiating the model
            training_args (TrainingArguments): Not used by this wrapper.
            preprocess_only (bool): If True, do not load model parameteres and only
                provide methods and attributes necessary for pre-processing the input data.
            **kwargs: passed to ``transformers.AutoModel.from_pretrained``.
        """
        super().__init__(args, **kwargs)

        if not preprocess_only:
            if self.args.use_peft or self.args.load_in_4bit or self.args.load_in_8bit:
                raise NotImplementedError

            torch_dtype = (
                self.args.torch_dtype
                if self.args.torch_dtype in ["auto", None]
                else getattr(torch, self.args.torch_dtype)
            )
            model_kwargs = dict(
                revision=self.args.model_revision,
                trust_remote_code=self.args.trust_remote_code,
                torch_dtype=torch_dtype,
            )
            self.model: PreTrainedModel = AutoModel.from_pretrained(
                self.args.model_name_or_path, **model_kwargs, **kwargs
            )
            # The wrapped model should take care of these
            # Expose these to remain compatible with a subset (and *NOT* all) of tools
            # (e.g., transformers.Trainer) that expect a transformers.PretrainedModel instance
            modeling_utils.add_model_apis_to_wrapper(wrapper=self, model=self.model)
        else:
            self.model = None

        self.append_eos_token = False

    def encode(self, inputs: Optional[Dict[str, torch.Tensor]]) -> torch.Tensor:
        outputs = self.model(**inputs)
        token_embeddings = outputs[0]
        mask = inputs["attention_mask"]
        token_embeddings = token_embeddings.masked_fill(~mask[..., None].bool(), 0.0)
        sentence_embeddings = token_embeddings.sum(dim=1) / mask.sum(dim=1)[..., None]
        return sentence_embeddings

    def format_query(self, text: str, **kwargs) -> str:
        return text.strip()

    def format_passage(self, text: str, title: Optional[str] = None, **kwargs) -> str:
        if title is None:
            return text.strip()
        else:
            return f"{title.strip()} {text.strip()}".strip()

    def save_pretrained(self, *args, **kwargs):
        if "state_dict" in kwargs:
            prefix = "model."
            assert all(k.startswith(prefix) for k in kwargs["state_dict"].keys())
            kwargs["state_dict"] = {
                k[len(prefix) :]: v for k, v in kwargs["state_dict"].items()
            }
        return self.model.save_pretrained(*args, **kwargs)



import os
from dataclasses import dataclass, fields
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import huggingface_hub as hf_hub


@dataclass
class MaterializedQRelConfig:
    """Information about a collection of queries, documents, and (optionally) the relation between
    them.

    You can use both path to local files and remote HF hub fsspec URIs for ``qrel_path``, ``corpus_path``, ``query_path``, and ``query_subset_path``.
    HF hub URIs must start with ``hf://``. See `HF hub documentation <https://huggingface.co/docs/huggingface_hub/en/guides/hf_file_system>`_ for the exact structure.

    If you do not set the value of ``qrel_path`` or set it to an empty list, ``MaterializedQRel`` will
    be a namespace container without any information about the relation between queries and
    documents. See :py:class:`trove.containers.materialized_qrel.MaterializedQRel` docstring for details.
    """

    qrel_path: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """One or multiple files that contain triplets of ``('qid', 'docid', 'score')``.
    The files do not need to explicitely contain such triplets. We can also infer them from other types of data.
    Look at ``file_reader.load_qrel()`` for supported files.
    """
    corpus_path: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """One or multiple files that contain the passage text and optionally titles.
    """
    query_path: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """One or multiple files that contain the query texts.
    """
    corpus_cache: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """(Not directly used) A Corresponding cache file name for each of the ``corpus_path`` to read/write the resulting embedding vectors.
    We do not directly use this. We just save it as part of the ``MaterializedQRel.args`` that you can use later.
    For example, you can use this to store a unique relative filepath for each of the corpus files.
    Then, during runtime, calculate a parent directory (e.g., based on embedding model name, etc.) and combine it with the relative filepath to get the complete path to cache files.
    """
    query_cache: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """(Not directly used) A Corresponding cache file name for each of the ``query_path`` to read/write the resulting embedding vectors.
    See docstring for ``corpus_cache`` for details.
    """
    query_subset_path: Optional[Union[os.PathLike, List[os.PathLike]]] = None
    """One or multiple files that it is possible to read a list of query IDs from.
    The available qrel triplets are limited to these queries. See ``file_reader.load_qids()`` for the supported files.
    """
    min_score: Optional[Union[int, float]] = None
    """If provided, filter the qrel triplets and only keep ones with ``min_score <= score`` (Endpoint is included in the interval)
    """
    max_score: Optional[Union[int, float]] = None
    """If provided, filter the qrel triplets and only keep ones with ``score < max_score`` (Endpoint is NOT included in the interval)
    """
    filter_fn: Optional[Callable[[Dict[str, Any]], bool]] = None
    """A callable used for filtering qrel triplets. If provided, ``min_score`` and ``max_score`` are ignored.
    `filter_fn` should take a dict (content of the qrel triplet with ``qid``, ``docid``, and ``score`` keys) as input and return a boolean as output.
    It is used like ``datasets.Dataset.filter(filter_fn, ...)``. I.e., keep the record if ``filter_fn`` returns ``True``.
    """
    group_top_k: Optional[int] = None
    """If given, filter the available documents for each query and only choose the ``group_top_k`` documents with the highest score for each query."""
    group_bottom_k: Optional[int] = None
    """If given, filter the available documents for each query and only choose the ``group_bottom_k`` documents with the lowest score for each query."""
    group_first_k: Optional[int] = None
    """If given, filter the available documents for each query and only keep the first ``group_first_k`` documents (in their original ordering) for each query."""
    group_random_k: Optional[int] = None
    """If given, filter the available documents for each query and choose ``group_random_k`` documents randomly for each query.
    Return all documents if number of available documents per query is smaller than ``group_random_k``.
    """
    group_filter_fn: Optional[
        Callable[[List[Dict[str, Any]]], List[Dict[str, Any]]]
    ] = None
    """A callable used to filter the qrel triplets for each query. If given, it overrides ``group_first_k``, ``group_top_k``, ``group_bottom_k``, and ``group_random_k``.
    There are several differences between ``group_filter_fn`` and ``filter_fn``.

        * ``filter_fn`` is used in the __init__ function to filter all the triplets for all queries and get the collection of available qrel triplets.
          But ``group_filter_fn`` is called whenever you attempt to get a list of available triplets for some query (i.e., every time you call methods like ``get_related_recs_for_*``).
          Unlike ``filter_fn``, results of ``group_filter_fn`` are not cached.

        * ``filter_fn`` operates on individual qrel triplets. But, ``group_filter_fn`` operates on the list of all available qrel triplets for some query.

    ``group_filter_fn`` must be a callable that takes one positional argument. The argument is a list of dict objects.
    Each dict object is a qrel triplet for the query. The dict object contains keys ``qid``, ``docid``, ``score``, and potentially other keys.
    The input list contains all the available qrel triplets for this query (the list could be empty).
    This callable should return an output with the same format as its input (i.e., a list of dicts).
    The behavior of this class is unknown if the callable receives a non-empty list but returns an empty list.
    If given, this callable is called before calling ``score_transform``.
    This argument is useful for filtering documents based on other documents available for each query.
    For example, to only keep the N most similar items for each query.
    """
    score_transform: Optional[
        Union[str, int, float, Callable[[Dict[str, Any]], Union[int, float]]]
    ] = None
    """A transformation applied to scores at the very last step right before returning them.
    Acceptable types for ``score_transform`` are:

    * ``None`` : return the scores as is
    * ``callable`` : it should take a dict (content of the qrel triplet with ``qid``, ``docid``, and ``score`` keys) as input and return the transformed score as output. it will be used like ``new_score = score_transform(rec)``
    * ``Union[int, float]`` : This value is used as the score for all qrel triplets. I.e., score is a constant for all query-document pairs
    * ``str`` : A predefined behavior. At the moment ``floor`` and ``ceil`` are valid behaviors. ``floor`` and ``ceil`` return ``int(triplet['score'])`` and ``math.ceil(triplet['score'])``, respectively
    """

    def __post_init__(self):
        self.ensure_list_of_correct_dtype()
        _non_empty_args = [
            k is not None
            for k in [
                self.group_top_k,
                self.group_bottom_k,
                self.group_first_k,
                self.group_random_k,
            ]
        ]
        if sum(_non_empty_args) > 1:
            msg = (
                "You should choose only one of the 'group_top_k', 'group_bottom_k', 'group_first_k', 'group_random_k' arguments."
                f" Got 'group_top_k': '{self.group_top_k}', 'group_bottom_k': '{self.group_bottom_k}', 'group_first_k': '{self.group_first_k}', and 'group_random_k': '{self.group_random_k}'"
            )
            raise ValueError(msg)

    def ensure_list_of_correct_dtype(self):
        """Ensure everything of type ``List[str]``."""
        # These three attrs must be lists. So, if they are 'None', that means an empty list
        if self.qrel_path is None:
            self.qrel_path = []
        if self.query_path is None:
            self.query_path = []
        if self.corpus_path is None:
            self.corpus_path = []

        # For consistency, make sure everything is of type `list` even if it is a single item
        if not isinstance(self.qrel_path, list):
            self.qrel_path = [self.qrel_path]

        # If query_cache is None and we have multiple query_path,
        # make sure there is one 'None' for each file
        if self.query_cache is None and isinstance(self.query_path, list):
            self.query_cache = [None] * len(self.query_path)

        # There should be one query_cache for each query_path
        assert isinstance(self.query_path, list) == isinstance(self.query_cache, list)

        # For consistency, make sure everything is a of type `list`
        if not isinstance(self.query_path, list):
            self.query_path = [self.query_path]
            self.query_cache = [self.query_cache]

        # There is a one-to-one correspondence between query_path and query_cache
        assert len(self.query_path) == len(self.query_cache)

        # Make sure everything is of type `str`
        self.query_path = [str(p) for p in self.query_path]
        self.query_cache = [None if p is None else str(p) for p in self.query_cache]

        # Repeat the same thing for corpus files
        if self.corpus_cache is None and isinstance(self.corpus_path, list):
            self.corpus_cache = [None] * len(self.corpus_path)
        assert isinstance(self.corpus_path, list) == isinstance(self.corpus_cache, list)
        if not isinstance(self.corpus_path, list):
            self.corpus_path = [self.corpus_path]
            self.corpus_cache = [self.corpus_cache]

        # There is a one-to-one correspondence between corpus_path and corpus_cache
        assert len(self.corpus_path) == len(self.corpus_cache)

        self.corpus_path = [str(p) for p in self.corpus_path]
        self.corpus_cache = [None if p is None else str(p) for p in self.corpus_cache]

        if self.query_subset_path is not None:
            if not isinstance(self.query_subset_path, list):
                self.query_subset_path = [self.query_subset_path]
            self.query_subset_path = list(sorted(self.query_subset_path))

        self.qrel_path = list(sorted(self.qrel_path))

        corpus_sorted_idx = sorted(
            range(len(self.corpus_path)), key=lambda i: self.corpus_path[i]
        )
        self.corpus_path = [self.corpus_path[i] for i in corpus_sorted_idx]
        self.corpus_cache = [self.corpus_cache[i] for i in corpus_sorted_idx]

        query_sorted_idx = sorted(
            range(len(self.query_path)), key=lambda i: self.query_path[i]
        )
        self.query_path = [self.query_path[i] for i in query_sorted_idx]
        self.query_cache = [self.query_cache[i] for i in query_sorted_idx]

        self._download_hf_hub_uri()

    def _download_hf_hub_uri(self):
        """Replaces huggingface hub URIs with path to downloaded local files.

        This method expects a python list for the following attributes: ``query_path``,
        ``corpus_path``, ``qrel_path``, and ``query_subset_path``.
        """
        fs = None

        def download_hf_hub_uri(path: Union[str, Path]) -> Union[str, Path]:
            # Download an hf hub URI and return its path
            if not isinstance(path, str) or not path.startswith("hf://"):
                return path

            nonlocal fs
            if fs is None:
                fs = hf_hub.HfFileSystem()

            info = fs.resolve_path(path)
            if info.path_in_repo == "":
                msg = f"The path must point to a file in the remote repo. Got: '{info.path_in_repo}'"
                raise ValueError(msg)

            local_path = hf_hub.hf_hub_download(
                repo_id=info.repo_id,
                repo_type=info.repo_type,
                revision=info.revision,
                filename=info.path_in_repo,
            )
            return local_path

        for path_list in [
            self.qrel_path,
            self.query_path,
            self.corpus_path,
            self.query_subset_path,
        ]:
            if path_list is None:
                continue
            for i in range(len(path_list)):
                path_list[i] = download_hf_hub_uri(path_list[i])

    def to_dict(self) -> Dict:
        """Return a json serializable view of the class attributes."""
        self.ensure_list_of_correct_dtype()
        json_dict = dict()
        field_names = [f.name for f in fields(self)]
        for fname in field_names:
            fvalue = getattr(self, fname)
            if fname in [
                "qrel_path",
                "corpus_path",
                "query_path",
                "corpus_cache",
                "query_cache",
                "query_subset_path",
            ]:
                if fvalue is not None:
                    # make sure paths are of type 'str'
                    fvalue = [v if v is None else str(v) for v in fvalue]
            if fname in ["filter_fn", "score_transform", "group_filter_fn"]:
                if fvalue is not None and callable(fvalue):
                    # for functions, just get their str representations
                    fvalue = str(fvalue)
            json_dict[fname] = fvalue
        return json_dict



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import pyarrow as pa
import torch

from .. import config, file_utils
from ..containers.key_to_row_indices import KeyToRowIndices
from ..logging_utils import get_logger_with_config, rpath

logger, logging_conf = get_logger_with_config("trove")


class VectorCacheMixin:
    def __init__(self, cache_file_name: Optional[os.PathLike] = None) -> None:
        """Mixin class for reading and writing dense vectors to pyarrow tables.

        This class provides the functionality to read and write dense vectors to pyarrow tables.
        It is intended to be used as a Mixin that provides this functionality to other classes with
        minimal effort.
        We use it to read and write query and passage embeddings to a cache file during
        evaluation of IR models.

        It only supports reading and writing ``(_id, value)`` tuples.

            * ``_id`` should be of type ``str``
            * ``value`` is a 1D numpy array of type `float32`
            * ``_id`` must uniquely identify each record in the entire dataset

        Args:
            cache_file_name (Optional[os.PathLike]): Path to cache file.
                Cache is disabled if not provided. It is also possible to add it after initialization.
        """
        # pyarrow schema representing a table with two columns,
        # '_id' (of type str) and 'value' (a 1D float vector)
        self.schema = pa.schema(
            [
                pa.field("_id", pa.string()),
                pa.field("value", pa.list_(pa.float32())),
            ]
        )

        self.cache_file_name: Optional[os.PathLike] = None
        # If true, we are allowed to write to cache file,
        # only allowed if cache file does not exist
        self.is_writable: Optional[bool] = None

        # pyarrow table holding cached values
        self.cache_table: Optional[pa.Table] = None
        # A mapping from _id to the corresponding row index in self.cache_table
        self._id_to_cache_idx: Optional[Union[Dict[str, int], KeyToRowIndices]] = None
        # Pyarrow file object used for writing
        self.file_pointer: Optional[pa.NativeFile] = None
        # pyarrow writer
        self.writer: Optional[pa.RecordBatchFileWriter] = None
        # Rows are written in batches. This holds rows that are waiting to be written to file.
        self.buffered_rows: List[Dict] = list()
        # is_writing is True if the cache file is open for writing
        self.is_writing: Optional[bool] = None

        # If not None, save the cache files in a subdir with this name in the directory where
        # cache files would've been saved otherwise.
        # Use this to save/load cache files after some permanent change has been made to the corresponding dataset.
        # This is not cleared when calling self.reset_state. It is permanent.
        # For example, if you do a _hard_ shard on the dataset, you should update this (by calling self.update_cache_subdir)
        # attribute to save the cache for hard sharded dataset in a separate subdirectory
        self.cache_variant_subdir: Optional[os.PathLike] = None

        self.update_cache_file_name(file_name=cache_file_name, load=False)

    @property
    def is_cache_available(self) -> bool:
        return self.cache_table is not None

    @property
    def effective_cache_file_name(self) -> Optional[os.PathLike]:
        """Effective cache file name after taking into account the ``cache_variant_subdir``
        property.

        You should always use this property to read/write cache files.
        """
        if self.cache_file_name is None:
            return None
        if self.cache_variant_subdir is None:
            return self.cache_file_name

        _fn = Path(self.cache_file_name)
        fname = _fn.parent.joinpath(self.cache_variant_subdir, _fn.name).as_posix()
        return fname

    def reset_state(self) -> None:
        """Resets the internal state of the class like a fresh instance without a cache filename.

        Does NOT change files on disk.
        """
        # Not allowed while writing to avoid corrupted cache files.
        assert (
            not self.is_writing
        ), "Attempting to reset the state while writing to the cache"

        self.cache_file_name = None
        self.is_writable = None
        self.cache_table = None
        if hasattr(self._id_to_cache_idx, "close"):
            self._id_to_cache_idx.close()
        self._id_to_cache_idx = None
        self.file_pointer = None
        self.writer = None
        self.buffered_rows = list()
        self.is_writing = None

    def unload_cache(self) -> None:
        """Purge the cache and index lookup tables that are loaded into memory.

        It reverses the impact of :meth:`load_cache`
        """
        # This will never be triggered because if the cache is loaded, it is not writable
        # But do it for consistency to emphasize for developers that no other operations
        # are allowed while writing to cache
        assert not self.is_writing, "Attempting to unload cache while writing"
        self.cache_table = None
        if hasattr(self._id_to_cache_idx, "close"):
            self._id_to_cache_idx.close()
        self._id_to_cache_idx = None

    def update_cache_subdir(
        self, subdir: Optional[os.PathLike], append: bool = True, load: bool = True
    ) -> None:
        """Update the cache nested subdir (e.g., after some permanent change to the corresponding
        dataset).

        When some permanent change has been made to the corresponding dataset (e.g., hard sharding),
        you should to let the ``VectorCache`` know where to save/load the cache files for the new dataset.
        This dataset also reset the state of loaded cache files, etc. (But does not change files on disk.).

        After updating cache subdir, it calls to :meth:`update_cache_file_name` initialize the new cache
        if it exists.

        Args:
            subdir (Optional[os.PathLike]): New cache files are saved in a subdirectory with this name
                in the parent directory that would've contained the original cache files.
            append (bool): overwrite or append to existing cache ``subdir`` if it exists.
            load (bool): If true, load the new cache file if it exists (the state of the previous cache tables
                is always cleared even if load is set to ``False``).
        """
        if append and subdir is None:
            msg = "cannot append `None` to existing `self.cache_variant_subdir`."
            raise TypeError(msg)

        # Keep track of the main cache filename
        # We are only changing the variant and not the main cache file name.
        prev_filename = self.cache_file_name
        if prev_filename is not None:
            prev_filename = Path(prev_filename).as_posix()

        if append:
            if self.cache_variant_subdir is None:
                self.cache_variant_subdir = Path(subdir).as_posix()
            else:
                self.cache_variant_subdir = Path(
                    self.cache_variant_subdir, subdir
                ).as_posix()
        else:
            msg = "Overwriting 'self.cache_variant_subdir' is not yet supported."
            raise NotImplementedError(msg)

        self.update_cache_file_name(file_name=prev_filename, load=load)

    def _get_vector_cache_shard_files(self) -> List[str]:
        """Returns a list of paths to cache shard files."""
        _cache_shard_files = list()
        if Path(self.effective_cache_file_name).parent.exists():
            # cache files can sometimes be sharded like 'file-000-of-001.arrow'
            # also exclude temporary incomplete files that were not removed
            _cache_shard_files = [
                p.absolute().as_posix()
                for p in Path(self.effective_cache_file_name).parent.rglob("*")
                if not p.is_dir()
                and p.name.startswith(Path(self.effective_cache_file_name).stem)
                and config.TROVE_INCOMPLETE_FILE_DIRNAME not in p.parts
            ]
        return sorted(_cache_shard_files)

    def update_cache_file_name(
        self, file_name: Optional[os.PathLike] = None, load: bool = True
    ) -> None:
        """Resets the state of the class and points to the new cache file.

        Args:
            file_name (Optional[os.PathLike]): New cache file.
            load (bool): If true and ``file_name`` is not ``None``, load the new cache file.
        """
        # Clear the old cache first
        self.reset_state()
        self.cache_file_name = file_name
        if self.effective_cache_file_name is None:
            # Nothing to do
            return

        _cache_shard_files = self._get_vector_cache_shard_files()
        # We are only allowed to write to this file
        # if it does not exist (and none of its sharded variants exist)
        self.is_writable = len(_cache_shard_files) == 0
        if load:
            self.load_cache()

    def load_cache(self) -> None:
        """Load the cache file as a memory mapped Arrow table."""
        # Writing should be finished (or not started at all) before loading the cache
        assert not self.is_writing, "Attempting the load the cache while writing"

        if self.cache_table is not None:
            # It is already loaded
            return

        if (
            self.effective_cache_file_name is None
            or not Path(self.effective_cache_file_name).parent.exists()
        ):
            # there is nothing to load
            return

        _cache_shard_files = self._get_vector_cache_shard_files()
        if not len(_cache_shard_files):
            # There is nothing to load
            return

        logger.info(
            f"Loading vector cache pyarrow table from {rpath(self.effective_cache_file_name)}. Num shards: {len(_cache_shard_files)}"
        )

        # load the cache files as memory mapped arrow tables
        cache_table = list()
        for file in _cache_shard_files:
            with pa.memory_map(file, "rb") as src:
                with pa.ipc.open_file(src) as reader:
                    cache_table.append(reader.read_all())

        if len(cache_table) == 1:
            cache_table = cache_table[0]
        else:
            cache_table = pa.concat_tables(cache_table)

        cached_ids = cache_table["_id"].to_numpy().tolist()
        assert len(cached_ids) == len(
            set(cached_ids)
        ), "Found duplicate record IDs in the cache"

        assert set(self.all_rec_ids) == set(
            cached_ids
        ), "Cached IDs are different from dataset IDs"

        logger.debug("Create cache id to row index.")
        # self._id_to_cache_idx = KeyToRowIndices(
        # dataset=datasets.Dataset.from_dict({"_id": cached_ids}),
        #     id_field="_id",
        #     num_proc=8,
        # )
        self._id_to_cache_idx = {_id: [_idx] for _idx, _id in enumerate(cached_ids)}
        self.cache_table = cache_table

    def get_cached_value(self, _id: str) -> Optional[np.ndarray]:
        """If possible, loads the cached value for the given ``_id``

        .. warning::

            The returned cached value shares memory with the arrow table and should not be modified in place.
            **Do not change this array in place.**

        Args:
            _id (str): Load the cached value corresponding to this unique ``_id``

        Returns:
            If cache exists, return the cached value as numpy array, otherwise, return ``None``.
        """
        if self.cache_table is None or self.is_writing:
            # You cannot read from cache while writing or if cache does not exist
            return None
        idx = self._id_to_cache_idx[_id][0]
        value = self.cache_table["value"][idx].values.to_numpy()
        return value

    @contextmanager
    def open_cache_io_streams(self):
        """A context manager to prepare for writing to cache files.

        It opens the cache file and creates the necessary write handlers.
        Should be used like::

            with instance.open_cache_io_streams():
                instance.cache_records(...) # You write to cache here
        """
        if self.effective_cache_file_name is None or not self.is_writable:
            # It is not possible to write to cache.
            # Return a dummy context manager to keep the user's code consistent whether they write to cache or not
            # We will raise an exception in other methods if the user tries to add records to cache when not possible
            yield
        else:
            assert (
                not self.is_writing
            ), "Attempting to write to file that is already open"
            # The shard information should be set by the class that inherits from this Mixin
            if (
                self.shard_idx is not None
                and self.shard_idx >= 0
                and self.num_shards is not None
                and self.num_shards > 0
            ):
                # If sharded creates files like '/path/to/cache/dir/cache_file-000-001.arrow'
                _pardir = Path(self.effective_cache_file_name).parent
                _cache_stem = Path(self.effective_cache_file_name).stem
                width = len(str(int(self.num_shards))) + 1
                suffix = f"-{self.shard_idx:0{width}}-of-{self.num_shards:0{width}}"
                sink_file_name = Path(_pardir).joinpath(
                    _cache_stem + suffix + Path(self.effective_cache_file_name).suffix
                )
            else:
                sink_file_name = Path(self.effective_cache_file_name)

            assert (
                not sink_file_name.exists()
            ), f"Cache file already exists:\n{sink_file_name.as_posix()}"
            sink_file_name.parent.mkdir(exist_ok=True, parents=True)

            logger.debug(
                f"Opened cache file for writing: {rpath(sink_file_name.as_posix())}"
            )
            with file_utils.atomic_write(file=sink_file_name, root="parent") as tfile:
                self.file_pointer = pa.OSFile(tfile.as_posix(), "wb")
                self.writer = pa.ipc.new_file(self.file_pointer, self.schema)
                self.is_writing = True
                # You can only write to a cache file once. No updating allowed.
                self.is_writable = False
                try:
                    yield
                finally:
                    # Write all the buffered records to file before closing the streams
                    self.flush()
                    self.writer.close()
                    self.file_pointer.close()
                    self.is_writing = False

    def cache_records(
        self,
        rec_id: List[str],
        value: Union[torch.Tensor, np.ndarray, List[np.ndarray]],
        chunk_size: int = 1_000,
    ) -> None:
        """Add a batch of records to the cache.

        The write is not immediate.
        It is written in chunks once a sufficient number of records are buffered.

        Args:
            rec_id (List[str]): a list of unique ``_id`` values
            value (Union[torch.Tensor, np.ndarray, List[np.ndarray]]): values to cache.
                It should either be a 2D ``torch.Tensor`` or 2D ``np.ndarray``.
                Each row is treated as a record to be written to the cache.
            chunk_size (int): The number of records to write at a time.
        """
        if self.effective_cache_file_name is None:
            # Do not raise exception if we are using the dataset without caching
            return

        # caching is enabled. Make sure we can write to cache file
        assert (
            self.is_writing
            and self.writer is not None
            and self.file_pointer is not None
            and self.file_pointer is not None
            and not self.file_pointer.closed
        )

        if isinstance(value, torch.Tensor):
            value = value.detach().cpu().numpy()

        if isinstance(value, np.ndarray):
            # Convert 2D array to a list of 1D arrays
            value = np.vsplit(value, value.shape[0])

        # Add to buffer and write to file later
        for _id, v in zip(rec_id, value):
            self.buffered_rows.append(
                {"_id": _id, "value": v.flatten().astype(np.float32)}
            )

        if len(self.buffered_rows) >= chunk_size:
            self.flush(chunk_size=chunk_size)

    def flush(self, chunk_size: int = 1_000) -> None:
        """Write buffered records to file.

        Args:
            chunk_size (int): The size of chunked arrays to write to arrow file.
        """
        # Make sure it is possible to write to cache file
        assert (
            self.is_writing
            and self.writer is not None
            and self.file_pointer is not None
            and not self.file_pointer.closed
            and self.file_pointer is not None
            and not self.file_pointer.closed
        )
        if len(self.buffered_rows):
            table = pa.Table.from_pylist(self.buffered_rows, schema=self.schema)
            self.writer.write_table(table, chunk_size)
            self.buffered_rows = []

    def _shard_checks_by_cache(self) -> None:
        """Do nothing if it is ok to shard the dataset, otherwise raise an exception."""
        assert (
            not self.is_writing
        ), "Attempting to shard the data while writing to cache file"

    @property
    def all_rec_ids(self) -> List[str]:
        """Returns the list of all record ids in the dataset across all shards."""
        raise NotImplementedError



from typing import Dict, Optional

import torch

from . import modeling_utils
from .pretrained_retriever import PretrainedRetriever, RetrieverOutput


class BiEncoderRetriever(PretrainedRetriever):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def forward(
        self,
        query: Optional[Dict[str, torch.Tensor]] = None,
        passage: Optional[Dict[str, torch.Tensor]] = None,
        label: Optional[torch.Tensor] = None,
        return_loss: Optional[bool] = True,
        **kwargs,
    ) -> RetrieverOutput:
        """Encodes query and passages and potentially calculate similarity scores and loss.

        Args:
            query (Optional[Dict[str, torch.Tensor]]): tokenized query.
            passage (Optional[Dict[str, torch.Tensor]]): tokenized passages.
            label (Optional[torch.Tensor]): Relevancy scores of the corresponding passages for each query.
                If there are k documents for each query, this is a 2D tensor of shape `[NUM_QUERIES, k]`
            return_loss (Optional[bool]): if true, calculate the loss value.
            **kwargs: unused keyword arguments are passed to the ``forward()`` method of the loss module.

        Returns:
            If possible, query and passage embeddings as well as similarity and loss scores.
        """
        # There is nothing to do if both query and passages are missing
        assert query is not None or passage is not None

        if query is not None:
            query = self.encoder.encode_query(inputs=query)

        if passage is not None:
            passage = self.encoder.encode_passage(inputs=passage)

        if query is None or passage is None:
            # if one of the passage or query is missing, we cannot do much beyond embedding them.
            return RetrieverOutput(query=query, passage=passage)

        if self.training and self.is_ddp:
            # Gather the query and passage embeddings from all the processes in a distributed environment.
            # After this if block, 'query' will contain the embedding of all the queries across all processes
            # Similarly, 'passage' will contain the embeddings of all passage from all processes
            # Also, 'label' is the value of input argument 'label' from all processes concatenated along dim 0
            query = modeling_utils.gather_tensors(
                tensor_obj=query, world_size=self.world_size, rank=self.process_rank
            )
            passage = modeling_utils.gather_tensors(
                tensor_obj=passage, world_size=self.world_size, rank=self.process_rank
            )
            label = modeling_utils.gather_tensors(
                tensor_obj=label, world_size=self.world_size, rank=self.process_rank
            )

        logits = self.similarity_fn(query=query, passage=passage)
        if not return_loss or self.loss is None:
            return RetrieverOutput(query=query, passage=passage, logits=logits)

        loss = self.loss(logits=logits, label=label, **kwargs)

        return RetrieverOutput(loss=loss, logits=logits, query=query, passage=passage)



from . import file_reader_functions
from .collator import RetrievalCollator
from .data_args import DataArguments
from .file_reader import (
    available_loaders,
    load_qids,
    load_qrel,
    load_records,
    register_loader,
)
from .file_transforms import (
    group_qrel_triplets_from_csv,
    group_qrel_triplets_from_sydir_corpus,
)
from .ir_dataset_binary import BinaryDataset
from .ir_dataset_multilevel import MultiLevelDataset
from .ir_encoding_dataset import EncodingDataset
from .vector_cache_mixin import VectorCacheMixin



import os
from dataclasses import asdict, is_dataclass
from typing import Any, Dict, List, Optional, Union

import wandb


class WandbCallback:
    def __init__(self) -> None:
        """Simple callback to log metrics to Wandb."""
        self.disabled = os.getenv("WANDB_DISABLED", "").upper() in {
            "1",
            "ON",
            "YES",
            "TRUE",
        }
        self.initialized = False
        self.wandb_run = None

    def setup(
        self,
        run_name: Optional[str] = None,
        args: Optional[Any] = None,
        model: Optional[Any] = None,
        init_kwargs: Optional[Dict] = None,
        extra_configs: Optional[Union[Dict, List[Dict]]] = None,
        **kwargs,
    ) -> None:
        """init a wandb run.

        Args:
            run_name (Optional[str]): human readable run name.
            args (Optional[Any]): instance of ``RetrievalEvaluator.args``
            model (Optional[Any]): model that is being evaluated
            init_kwargs (Optional[Dict]): additional kwargs for ``wandb.init()`` method
            extra_configs (Optional[Union[Dict, List[Dict]]]): additional configs to be merged with experiment config when initializing wandb.
            **kwargs: Not used. Just to swallow extra arguments.
        """
        if self.disabled or self.initialized:
            return
        self.initialized = True

        if init_kwargs is None:
            init_kwargs = dict()
        if "name" not in init_kwargs and run_name is not None:
            init_kwargs["name"] = run_name

        if "project" not in init_kwargs:
            init_kwargs["project"] = os.getenv("WANDB_PROJECT", "trove")

        if args is None:
            init_config = dict()
        else:
            init_config = {**args.to_dict()}

        if model is not None:
            if hasattr(model, "args") and model.args is not None:
                if isinstance(model.args, dict):
                    init_config = {**model.args, **init_config}
                elif hasattr(model.args, "to_dict"):
                    init_config = {**model.args.to_dict(), **init_config}
                else:
                    if is_dataclass(model.args):
                        args_dict = asdict(model.args)
                        init_config = {**args_dict, **init_config}
            elif hasattr(model, "config") and model.config is not None:
                model_config = (
                    model.config
                    if isinstance(model.config, dict)
                    else model.config.to_dict()
                )
                init_config = {**model_config, **init_config}

        if extra_configs is not None:
            if isinstance(extra_configs, dict):
                init_config = {**extra_configs, **init_config}
            elif isinstance(extra_configs, list):
                for conf in extra_configs:
                    init_config = {**conf, **init_config}
            else:
                msg = f"tracker_extra_configs should either be a dict or a list of dicts. Got: {type(extra_configs)}"
                raise TypeError(msg)

        self.wandb_run = wandb.init(**init_kwargs, config=init_config)

    def log(self, *args, **kwargs):
        """Log results to wandb run."""
        if self.disabled:
            return
        return self.wandb_run.log(*args, **kwargs)



import copy
import itertools
import math
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import datasets
import numpy as np
from torch.utils.data import Dataset

from ..logging_utils import get_logger_with_config
from .data_args import DataArguments
from .vector_cache_mixin import VectorCacheMixin

logger, logging_conf = get_logger_with_config("trove")


class EncodingDataset(VectorCacheMixin, Dataset):
    def __init__(
        self,
        data_args: Optional[DataArguments] = None,
        data_args_overrides: Optional[Dict[str, Any]] = None,
        dataset_name: Optional[str] = None,
        query_path: Optional[os.PathLike] = None,
        corpus_path: Optional[os.PathLike] = None,
        format_query: Optional[Callable[[str, Optional[str]], str]] = None,
        format_passage: Optional[Callable[[str, str, Optional[str]], str]] = None,
        global_id_suffix: Optional[str] = None,
        cache_path: Optional[os.PathLike] = None,
        load_cache_on_init: bool = True,
        prefer_cache: bool = True,
    ) -> None:
        """Dataset for encoding query and passage texts.

        **This class supports reading/writing dense embedding vectors to cache on disk.**

        Args:
            data_args (Optional[DataArguments]): general arguments for loading and processing the data.
                Currently only used to find out ``dataset_name`` if not explicitely provided.
                And that is only useful if your query/passage formatting functions require a ``dataset`` argument.
            data_args_overrides (Optional[Dict[str, Any]]): A mapping from a subset of ``DataArguments`` attribute names to
                their new values. These key values override the corresponding attributes in ``data_args`` argument.
                It is useful if you want to create multiple datasets from the same ``DataArguments`` instance
                but make small changes for each dataset without creating new ``DataArguments`` instances.
            dataset_name (Optional[str]): Name of the dataset that is being encoded.
                only useful if your query/passage formatting functions require a ``dataset`` argument.
            query_path (Optional[os.PathLike]): Path to JSONL file containing query texts.
            corpus_path (Optional[os.PathLike]): Path to JSONL file containing passage texts.
            format_query (Optional[Callable]): callable that takes query text and dataset name and returns the formatted query text.
            format_passage (Optional[Callable]): callable that takes passage text and title and dataset name and returns the formatted passage text.
            global_id_suffix (Optional[str]): unique file id used to create globally unique query/passage ids across files.
            cache_path (Optional[os.PathLike]): path to cache file
            load_cache_on_init (bool): if true and ``cache_path`` is provided and exists, load the cache arrow table at the end of init method
            prefer_cache (bool): If true, use cache if present. If false, ignore cache even if it exists.
        """
        super().__init__(cache_file_name=cache_path)

        # We only support one source of data
        if ((query_path is None) == (corpus_path is None)) or (
            (format_query is None) == (format_passage is None)
        ):
            msg = (
                "We only support encoding either queries or documents in each instance of 'EncodingDataset'."
                " You should provide one and only one of the ('query_path', 'format_query')"
                " or ('corpus_path', 'format_passage') argument pairs."
                f" Got: 'query_path': '{query_path}', 'format_query': '{format_query}',"
                f" 'corpus_path': '{corpus_path}', 'format_passage': '{format_passage}'"
            )
            raise ValueError(msg)

        # make a deepcopy so we do not impact operations outside of this class.
        data_args = copy.deepcopy(data_args)
        if data_args_overrides is not None:
            for attr, val in data_args_overrides.items():
                if not hasattr(data_args, attr):
                    # We should not add new attributes
                    msg = f"'{attr}' is not a valid attribute name for 'DataArguments'"
                    raise ValueError(msg)
                setattr(data_args, attr, val)
        self.args = data_args

        self.dataset_name = None
        if dataset_name is not None:
            self.dataset_name = dataset_name
        elif self.args is not None and self.args.dataset_name is not None:
            self.dataset_name = self.args.dataset_name

        self._prefer_cache = prefer_cache
        self.format_query = format_query
        self.format_passage = format_passage

        # See `MaterializedQRel.local_to_global_id()` for more info
        if global_id_suffix is not None:
            self.global_id_suffix = "_" + global_id_suffix
        else:
            self.global_id_suffix = ""

        if query_path is not None:
            self.filepath = query_path
            self.is_query = True

        if corpus_path is not None:
            self.filepath = corpus_path
            self.is_query = False

        if Path(self.filepath).suffix not in [".json", ".jsonl"]:
            msg = "Only json/jsonl files are supported"
            raise NotImplementedError(msg)

        logger.info(f"Load Encoding data from file:\n{self.filepath}")
        self.dataset = datasets.load_dataset(
            "json", data_files=self.filepath, split="train"
        )
        # Index of all rows that this dataset can access.
        self.all_row_indices = list(range(len(self.dataset)))
        # Index of rows in the current shard.
        self.shard_row_indices = list(range(len(self.dataset)))
        self.shard_idx = None
        self.num_shards = None

        # Find the field that contains the query/passage ID
        ds_fields = list(self.dataset[0].keys())
        all_id_fields = [
            "_id",
            "qid",
            "query_id",
            "query-id",
            "docid",
            "passage_id",
            "passage-id",
            "corpus-id",
        ]
        ds_id_fields = list(set(ds_fields).intersection(set(all_id_fields)))
        assert len(ds_id_fields) == 1
        self.id_field = ds_id_fields[0]

        # Unique ID of all the available records in the entire dataset regardless of soft sharding
        self._all_rec_ids = self.dataset[self.id_field]

        if load_cache_on_init:
            self.load_cache()

    @property
    def all_rec_ids(self) -> List[str]:
        return self._all_rec_ids

    def ignore_cache(self) -> None:
        """Return raw data even if cache exists."""
        self._prefer_cache = False

    def prefer_cache(self) -> None:
        """Return cached embeddings if possible."""
        self._prefer_cache = True

    @contextmanager
    def disable_cache(self):
        """Context manager to temporarily disable vector cache."""
        orig_state = self._prefer_cache
        self.ignore_cache()
        try:
            yield
        finally:
            self._prefer_cache = orig_state

    def shard(
        self,
        shard_idx: int,
        num_shards: Optional[int] = None,
        shard_weights: Optional[List[float]] = None,
        hard_shard: bool = False,
    ) -> None:
        """Shard the dataset.

        It can do both a `hard` and `soft` shard.

        In a soft shard, we only mask the index of rows that are not in the shard.
        And, the underlying data is not changed and you can reverse the sharding by calling the ``unshard()`` method.

        In a hard shard, however, the underlying data is changed and the effects of sharding are irreversible for this instance.
        Also keep in mind that after a hard shard, the dataset will seem unsharded to other methods/classes like ``VectorCacheMixin``.

        In most cases, you should just use the default soft sharding.
        Hard sharding is mostly useful if you plan to shard the dataset twice.
        For example, if you want to encode each shard in a separate session which itself runs in a distributed environment,
        You can do a hard shard in each session before passing the dataset to ``trove.RetrievalEvaluator()``, for instance.
        This allows ``RetrievalEvaluator`` to further shard the dataset in each session into smaller pieces for each process.

        Args:
            shard_idx (int): Index of current shards
            num_shards (Optional[int]): Total number of shards. If ``shard_weights`` is
                provided, ``num_shards`` should be either ``None`` or equal to ``len(shard_weights)``.
            shard_weights (Optional[List[float]]): relative number of items in each shard.
                If provided, shard the dataset according to these weights. If not provided, all shards
                are of the same size.
            hard_shard (bool): Whether to a hard/permanent shard or a soft/superficial shard.
        """
        if shard_weights is None and num_shards is None:
            msg = "You must provide at least one of 'shard_weights' or 'num_shards'. Got None for both."
            raise ValueError(msg)
        if (
            shard_weights is not None
            and num_shards is not None
            and len(shard_weights) != num_shards
        ):
            msg = "'num_shards' is given but is not equal to 'len(shard_weights)'"
            raise ValueError(msg)

        if self.shard_idx is not None or self.num_shards is not None:
            msg = (
                "Nested soft sharding is not supported."
                " Nested sharding leads to similar cache file names for different subsets of data."
                " You might want to do a hard shard followed by a soft shard to achieve similar results."
            )
            raise NotImplementedError(msg)

        # Make sure it is ok to shard the dataset from the cache perspective.
        # E.g. You cannot shard dataset if cache file is open for writing.
        self._shard_checks_by_cache()

        num_items = len(self.all_row_indices)
        if shard_weights is None:
            # Chunk the dataset into shards of equal sizes
            shard_size = math.ceil(num_items / num_shards)
            shard_boundaries = [shard_idx * shard_size, (shard_idx + 1) * shard_size]
        else:
            # The size of each shard is determined by shard_weights
            num_shards = len(shard_weights)
            norm_shard_weights = [i / sum(shard_weights) for i in shard_weights]
            shard_sizes = [math.ceil(num_items * i) for i in norm_shard_weights]
            all_shard_boundaries = list(itertools.accumulate([0] + shard_sizes))
            assert all_shard_boundaries[-1] >= num_items
            all_shard_boundaries = list(itertools.pairwise(all_shard_boundaries))
            shard_boundaries = all_shard_boundaries[shard_idx]

        self.shard_row_indices = self.all_row_indices[
            shard_boundaries[0] : shard_boundaries[1]
        ]

        self.shard_idx = shard_idx
        self.num_shards = num_shards

        if hard_shard:
            # Forget everything about other shards.
            # Make it look like this shard is the only thing in the dataset
            # Make sure to use shallow copy when updating indices.
            self.all_row_indices = self.shard_row_indices[:]
            self._all_rec_ids = self.dataset.select(self.all_row_indices)[self.id_field]
            # Since in a hard shard, the dataset looks as unsharded to other classes including `VectorCacheMixin`,
            # We have to manually separate the cache file for this shard from that of other shards.
            # We do it by saving all vector cache files related to this shard in a separate subdirectory
            # in the directory that unsharded cache files would have been saved
            width = len(str(int(self.num_shards))) + 1
            cache_subdir = (
                f"hard_shard-{self.shard_idx:0{width}}-of-{self.num_shards:0{width}}"
            )
            # Make it look like unsharded dataset
            self.shard_idx = None
            self.num_shards = None
            # Let VectorCacheMixin know that something permanent has changed about this dataset
            self.update_cache_subdir(subdir=cache_subdir, append=True, load=False)

    def unshard(self) -> None:
        """Reverse the sharding.

        I.e., unmask the index of rows accessible to dataset
        """
        # Make a shallow copy and do not assign directly
        self.shard_row_indices = self.all_row_indices[:]
        self.shard_idx = None
        self.num_shards = None

    def __len__(self) -> int:
        # Just consider the rows that we are allowed to access
        return len(self.shard_row_indices)

    def __getitem__(self, index: int) -> Dict[str, Union[str, np.ndarray]]:
        """Return either the formatted raw data or cached embeddings."""
        rec = self.dataset[self.shard_row_indices[index]]
        rec_id = rec[self.id_field]  # Unique query/document ID

        output = dict()
        emb = None
        if self._prefer_cache:
            emb = self.get_cached_value(rec_id)

        if emb is not None:
            # only return cached value if found/exists
            output["embedding"] = emb
        else:
            if self.is_query:
                output["query"] = self.format_query(
                    text=rec["text"], dataset=self.dataset_name
                )
            else:
                output["passage"] = self.format_passage(
                    text=rec["text"], title=rec["title"], dataset=self.dataset_name
                )

        # The original record ID and the global record ID used to allow
        # merging these records with records from other files.
        # See `MaterializedQRel.local_to_global_id()` for more info
        # **You should always use `orig_rec_id` to read/write from/to cache.
        output["orig_rec_id"] = rec_id
        output["rec_id"] = rec_id + self.global_id_suffix

        return output



"""Data collator for information retrieval."""

import itertools
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import numpy as np
import torch
from transformers import PreTrainedTokenizer

from ..logging_utils import get_logger_with_config
from .data_args import DataArguments

logger, logging_conf = get_logger_with_config("trove")


def get_tokenized(
    inputs: List[str],
    tokenizer: PreTrainedTokenizer,
    max_length: int,
    append_eos: bool,
    pad_to_multiple_of: Optional[int],
) -> Dict[str, torch.Tensor]:
    """Tokenize the input texts.

    From here:
    https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/collator.py

    Args:
        inputs (List[str]): list of input texts.
        tokenizer (PreTrainedTokenizer): transformers tokenizer
        max_length (int): max length of the result
        append_eos (bool): whether to append the ``eos`` token to ``input_ids``
        pad_to_multiple_of (Optional[int]): ``pad_to_multiple_of`` argument of the tokenizer

    Returns:
        tokenized inputs in pytorch tensors.
    """

    output = tokenizer(
        inputs,
        padding=False,
        truncation=True,
        max_length=max_length - 1 if append_eos else max_length,
        return_attention_mask=False,
        return_token_type_ids=False,
        add_special_tokens=True,
    )

    if append_eos:
        output["input_ids"] = [
            i + [tokenizer.eos_token_id] for i in output["input_ids"]
        ]

    output = tokenizer.pad(
        output,
        padding=True,
        pad_to_multiple_of=pad_to_multiple_of,
        return_attention_mask=True,
        return_tensors="pt",
    )
    return output


@dataclass
class RetrievalCollator:
    data_args: DataArguments
    tokenizer: PreTrainedTokenizer
    append_eos: bool

    def __post_init__(self):
        # keys are names of the fields in the output of the dataset (e.g., passage, label, etc.)
        # values are a boolean that is 'True' if the field can be converted to a torch tensor and false otherwise
        # When we face a field that we do not recognize, we first try to convert it to a tensor and if we
        # succeed, we set the corresponding value to 'True' and false otherwise.
        # the next time that we encounter the field, it is faster to decide whether it should be converted to tensor or not
        self.field_is_tensor = dict()

    def __call__(
        self, features: List[Dict[str, Union[str, List[str], List[Union[int, float]]]]]
    ) -> Dict[str, Union[torch.Tensor, Dict[str, torch.Tensor]]]:
        """tokenize and collate a list of examples.

        Args:
            features (List[Dict[str, Union[str, List[str], List[int]]]]): A list of examples (i.e., outputs of the dataset `__getitem__` function).
                We only process these keys in each example dictionary: ``['query', 'passage', 'label']``

                * ``'query'``: a text string
                * ``'passage'``: a list of text strings
                * ``'label'``: a list of integers or floats

                If there are additional keys, their values are collected in a list and returned as is.

        Returns:
            a mapping with collated ``'query'`` and ``'passage'`` tokens if available.
            If available in input features, ``'label'`` are also returned as a torch tensor.
        """
        features_dict = defaultdict(list)

        for example in features:
            for key, value in example.items():
                features_dict[key].append(value)

        output = dict()

        if len(features_dict["query"]) != 0:
            output["query"] = get_tokenized(
                inputs=features_dict["query"],
                tokenizer=self.tokenizer,
                max_length=self.data_args.query_max_len,
                append_eos=self.append_eos,
                pad_to_multiple_of=self.data_args.pad_to_multiple_of,
            )
        if len(features_dict["passage"]) != 0:
            if isinstance(features_dict["passage"][0], list):
                features_dict["passage"] = list(
                    itertools.chain.from_iterable(features_dict["passage"])
                )
            output["passage"] = get_tokenized(
                inputs=features_dict["passage"],
                tokenizer=self.tokenizer,
                max_length=self.data_args.passage_max_len,
                append_eos=self.append_eos,
                pad_to_multiple_of=self.data_args.pad_to_multiple_of,
            )
        if len(features_dict["embedding"]) != 0:
            output["embedding"] = torch.tensor(np.array(features_dict["embedding"]))
        if len(features_dict["label"]) != 0:
            output["label"] = torch.tensor(features_dict["label"])

        _already_checked = ["query", "passage", "label", "embedding"]
        for feat_name, feat_values in features_dict.items():
            if feat_name in _already_checked or len(feat_values) == 0:
                continue
            if feat_name not in self.field_is_tensor:
                try:
                    torch.tensor(feat_values)
                    self.field_is_tensor[feat_name] = True
                except:
                    self.field_is_tensor[feat_name] = False
            if self.field_is_tensor[feat_name]:
                output[feat_name] = torch.tensor(feat_values)
            else:
                output[feat_name] = feat_values

        return output



from . import config, data, modeling
from .containers import (
    FastResultHeapq,
    MaterializedQRel,
    MaterializedQRelConfig,
    ResultHeapq,
    RowsByKey,
    RowsByKeySingleSource,
)
from .data import (
    BinaryDataset,
    DataArguments,
    EncodingDataset,
    MultiLevelDataset,
    RetrievalCollator,
    available_loaders,
    register_loader,
)
from .evaluation import (
    EvaluationArguments,
    IRMetrics,
    RelevanceEvaluatorPlus,
    RetrievalEvaluator,
)
from .modeling import (
    BiEncoderRetriever,
    ModelArguments,
    PretrainedEncoder,
    PretrainedRetriever,
    RetrievalLoss,
    RetrieverOutput,
)
from .trainer import RetrievalTrainer, RetrievalTrainingArguments

__version__ = "0.0.1"



from collections import defaultdict
from typing import Dict, List, Set, Union

import pytrec_eval

from .mrr import MRRCutEvaluator


class RelevanceEvaluatorPlus:
    def __init__(
        self,
        query_relevance: Dict[str, Dict[str, Dict[str, int]]],
        measures: Union[Set[str], List[str]],
        relevance_level: int = 1,
    ) -> None:
        """Expands ``pytrec_eval.RelevanceEvaluator`` with new metrics.

        User facing API is the same as ``pytrec_eval.RelevanceEvaluator`` and should be a drop-in replacement.
        See ``pytrec_eval.RelevanceEvaluator`` for more details.


        Args:
            query_relevance (Dict[str, Dict[str, Dict[str, int]]]): see ``pytrec_eval.RelevanceEvaluator.__init__()``
            measures (Union[Set[str], List[str]]): see ``pytrec_eval.RelevanceEvaluator.__init__()``
            relevance_level (int): see ``pytrec_eval.RelevanceEvaluator.__init__()``
        """
        _new_evaluators = [MRRCutEvaluator]
        basename_to_evaluator = {_cls.measure_name: _cls for _cls in _new_evaluators}
        # Requested measures for new Evaluators
        basename_to_measures = {k: set() for k in basename_to_evaluator.keys()}

        pytrec_measures = set()
        # We assume none of the new Evaluators define the same base measure name as the ones supported by pytrec_eval.
        for measure in measures:
            for bname in basename_to_measures.keys():
                if measure.startswith(bname):
                    basename_to_measures[bname].add(measure)
                    break
            else:
                # We assume pytrec_eval supports the measure if it is not supported by any of the new Evaluators
                pytrec_measures.add(measure)

        if len(pytrec_measures) != 0:
            self.pytrec_evaluator = pytrec_eval.RelevanceEvaluator(
                query_relevance=query_relevance,
                measures=pytrec_measures,
                relevance_level=relevance_level,
            )
        else:
            self.pytrec_evaluator = None

        self.other_evaluators = list()
        for other_bname, other_measures in basename_to_measures.items():
            if len(other_measures) != 0:
                other_eval_cls = basename_to_evaluator[other_bname]
                self.other_evaluators.append(
                    other_eval_cls(
                        query_relevance=query_relevance,
                        measures=other_measures,
                        relevance_level=relevance_level,
                    )
                )

    def evaluate(
        self, scores: Dict[str, Dict[str, Dict[str, float]]]
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """Calculate IR metrics for the given similarity scores.

        User facing API is the same as ``pytrec_eval.RelevanceEvaluator.evaluate()`` and should be a drop-in replacement.
        See ``pytrec_eval.RelevanceEvaluator.evaluate()`` for more details.

        Args:
            scores (Dict[str, Dict[str, Dict[str, float]]]): predicted similarity scores.
                see ``pytrec_eval.RelevanceEvaluator.evaluate()``
        Returns:
            a mapping where ``output[qid][metric_name]`` is the value of "metric_name" for query "qid".
            see ``pytrec_eval.RelevanceEvaluator.evaluate()``
        """
        if self.pytrec_evaluator is not None:
            metrics = self.pytrec_evaluator.evaluate(scores)
        else:
            metrics = defaultdict()

        for other_eval in self.other_evaluators:
            other_metrics = other_eval.evaluate(scores)
            # merge with other metrics already calculated
            for qid, qmetrics in other_metrics.items():
                metrics[qid] = {**metrics[qid], **qmetrics}
        if isinstance(metrics, defaultdict):
            metrics = dict(metrics)
        return metrics



from typing import Dict, List, Set, Union

from . import metric_utils


class MRRCutEvaluator:
    measure_name: str = "mrr_cut"

    def __init__(
        self,
        query_relevance: Dict[str, Dict[str, Dict[str, int]]],
        measures: Union[str, List[str], Set[str]],
        relevance_level: int = 1,
    ) -> None:
        """Calculate MRR with cutoff values.

        The basename for this metric is ``mrr_cut``. The user facing API mimics that of ``pytrec_eval.RelevanceEvaluator``.

        Args:
            query_relevance (Dict[str, Dict[str, Dict[str, int]]]): See ``pytrec_eval.RelevanceEvaluator.__init__``.
            measures (Union[str, List[str], Set[str]]): See ``pytrec_eval.RelevanceEvaluator.__init__``.
            relevance_level (int): docs with ground truth relevancy level ``>= relevance_level`` are considered relevant.
                See ``pytrec_eval.RelevanceEvaluator.__init__``.
        """
        measures = metric_utils.expand_and_combine_eval_measures(
            measures=measures,
            supported_measures=self.__class__.measure_name,
            reconstruct=False,
        )
        if len(measures[self.__class__.measure_name]) == 0:
            msg = "There are no default cutoff values for 'mrr_cut'. Please specify cutoff values like 'mrr_cut.k1,k2,k3'."
            raise ValueError(msg)
        k_values = list(measures[self.__class__.measure_name])
        for k in k_values:
            if int(k) != float(k):
                msg = f"cutoff values for 'mrr_cut' must be integers. Got '{k}'"
                raise ValueError(msg)
        self.k_values = [int(k) for k in k_values]

        # docs for each query that are considered related according to groundtruth qrels.
        self.related_docs = {
            qid: {_id for _id, _sc in qdata.items() if _sc >= relevance_level}
            for qid, qdata in query_relevance.items()
        }

    def evaluate(
        self, scores: Dict[str, Dict[str, Dict[str, float]]]
    ) -> Dict[str, Dict[str, Dict[str, float]]]:
        """Calculate MRR for the given similarity scores.

        Same API as ``pytrec_eval.RelevanceEvaluator.evaluate()``

        Args:
            scores (Dict[str, Dict[str, Dict[str, float]]]): predicted similarity scores.
                Same as args for ``pytrec_eval.RelevanceEvaluator.evaluate()``.

        Returns:
            Calculated MRR values. ``ret[qid][mrr_cut_k]`` is the "MRR@k" for qid.
            The output structure is the same as ``pytrec_eval.RelevanceEvaluator.evaluate()``.
        """
        max_k = max(self.k_values)
        common_qids = set(self.related_docs).intersection(scores)

        query_mrrs = dict()  # holds MRR values for each query
        for qid in common_qids:
            qdata = scores[qid]
            sorted_docids = sorted(qdata, key=qdata.get, reverse=True)[:max_k]

            # 'rank' is the rank of the first related document for the query
            rank = max_k + 1
            for r, _id in enumerate(sorted_docids):
                if _id in self.related_docs[qid]:
                    rank = r + 1
                    # we are only interested in the rank of the first related document
                    break

            query_mrrs[qid] = dict()
            for k in self.k_values:
                m_name = f"{self.__class__.measure_name}_{k}"
                if k < rank:
                    query_mrrs[qid][m_name] = 0
                else:
                    query_mrrs[qid][m_name] = 1.0 / rank

        return query_mrrs



"""Some data transformations with polars that take a file as input and write to an output file.

These are just concerned with the main computation. Caching and other tasks should be done by the
user of these functions.
"""

import functools
import json
import os
import pickle as pkl
import sys
from pathlib import Path
from subprocess import run
from typing import Any, Dict


def call_script_fn(fn_name: str, kwargs: Dict[str, str]) -> None:
    """Call the file that the function is defined in as the main module with specific arguments.

    if you call ``call_script_fn('myfn', {'foo': 'bar'})``, this function calls the following shell function: ``$ python CURRENT_FILE myfn --foo bar``

    Together with the code in ``__main__`` of this file, this function allows us to call functions from this module
    in a separate process without using multiprocessing (e.g., fork). We want to avoid forking the
    process because polars has some problems with that in specific python versions.

    In general, we want to run polars functions in a separate process so the OS frees all the memory once the
    process is over. In early versions of polars, the memory was not released even after polar operations were done.
    This trick helps with that.

    Args:
        fn_name: the name of the function to call.
        kwargs: the keyword arguments to the target function.
    """
    cmd = list()
    cmd.append(Path(sys.executable).absolute().as_posix())
    cmd.append(Path(__file__).absolute().as_posix())
    cmd.append(fn_name)
    for k, v in kwargs.items():
        cmd.append(f"--{k}")
        cmd.append(f"{v}")
    run(cmd, check=True, cwd=Path.cwd().absolute().as_posix())


def csv_loader_kwargs(filepath: os.PathLike) -> Dict[str, Any]:
    """Automatically detect appropriate kwargs for ``pd.read_csv`` and similar tools.

    Args:
        filepath (os.PathLike): Path to delimiter separate file.

    Returns:
        A kwargs dict to be used like ``pd.read_csv(**kwargs)``
    """
    kwargs = dict(float_precision="high")

    # Inspect five lines to be more confident
    lines = list()
    with open(filepath, "r") as f:
        for i, line in enumerate(f):
            lines.append(line)
            if i > 10:
                break

    # Check for common delimiters
    if all("\t" in line.strip() for line in lines):
        sep = "\t"
    elif all("," in line.strip() for line in lines):
        sep = ","
    elif all(" " in line.strip() for line in lines):
        sep = " "
    else:
        raise RuntimeError
    kwargs["sep"] = sep

    cols = [c.strip() for c in lines[0].strip().split(sep)]

    if len(cols) != 3:
        msg = "We only support delimiter separated files with three columns in this order (with or without headers): 'qid', 'docid', 'score'"
        raise NotImplementedError(msg)

    # common header names in IR csv/tsv files
    header_name_mapping = {
        "qid": "qid",
        "query_id": "qid",
        "query-id": "qid",
        "docid": "docid",
        "passage_id": "docid",
        "passage-id": "docid",
        "corpus-id": "docid",
        "score": "score",
    }

    matching_header_names = list(
        set(list(header_name_mapping.keys())).intersection(set(cols))
    )
    if len(matching_header_names) == 0:
        kwargs["header"] = None
    elif len(matching_header_names) == 3:
        kwargs["header"] = 0
    else:
        raise RuntimeError

    if kwargs["header"] is None:
        # This is the only supported format at the moment.
        kwargs["names"] = ["qid", "docid", "score"]
    else:
        kwargs["names"] = [header_name_mapping[i] for i in cols]

    return kwargs


def _key_to_row_indices_mapping_op(
    src_file: os.PathLike, dst_file: os.PathLike, id_field: str
) -> None:
    """Create key to row indices mapping using polars.

    Given a source file and a column name (``id_field``), this function creates a mapping from the values of the given column
    to the index of rows that contain that value (In fact, it is a list of row indices).

    ``src_file`` should be a JSONL file or an Arrow IPC file that contains a field named ``id_field``.
    The result written to ``dst_file`` is a pickle containing a dict from values of key field
    to list of indices of rows that contain those values.

    Args:
        src_file (os.PathLike): JSONL or arrow IPC file with key for each row
        dst_file (os.PathLike): a pickle file containing a dict from values of key field
            to list of indices of rows that contain those values.
        id_field (str): The name of the column that should be used as key
    """
    # logger.debug("Running Polars query to create key to row indices mapping.")

    import polars as pl

    if Path(src_file).suffix not in [".jsonl", ".arrow"]:
        msg = f"Expected a json lines or arrow IPC file (with '.jsonl' or '.arrow' suffix) as input. Got '{src_file}'"
        raise ValueError(msg)

    dst_file = Path(dst_file)
    if dst_file.suffix not in [".pkl", ".pickle"]:
        msg = f"We only support writing to pickle file. But, got: {dst_file.as_posix()}"
        raise ValueError(msg)

    if Path(src_file).suffix == ".jsonl":
        # check if the file is empty or has records
        empty_file = True
        with open(src_file, "r") as f:
            for line in f:
                if line.strip() != "":
                    empty_file = False
                    break

        if empty_file:
            # avoid errors if the file is empty
            rows_by_key = dict()
        else:
            df = pl.scan_ndjson(src_file)
            df = df.select(pl.col(id_field)).with_row_index()
            rows_by_key = df.collect().rows_by_key(id_field)
    elif Path(src_file).suffix == ".arrow":
        df = pl.scan_ipc(src_file, cache=False)
        df = df.select(pl.col(id_field)).with_row_index()
        rows_by_key = df.collect().rows_by_key(id_field)
    else:
        raise ValueError

    # in earlier versions of polars, if df had only one column, the output of polars was
    # a list of indices. In later versions, the output is a tuple of length one in such cases.
    # If it is a tuple, we need to flatten it before writing it to disk
    _sample_idx = next(iter(rows_by_key.values()))[0]
    if isinstance(_sample_idx, tuple):
        idx_is_tuple = True
    else:
        idx_is_tuple = False

    sorted_keys = sorted(list(rows_by_key.keys()), key=lambda x: str(x))
    sorted_dict = dict()
    for key in sorted_keys:
        _vals = rows_by_key[key]
        if idx_is_tuple:
            _vals = [i[0] for i in _vals]
        sorted_dict[str(key)] = sorted(_vals)

    dst_file.parent.mkdir(exist_ok=True, parents=True)
    with open(dst_file, "wb") as f:
        pkl.dump(sorted_dict, f)


@functools.wraps(_key_to_row_indices_mapping_op)
def key_to_row_indices_mapping_op(
    src_file: os.PathLike, dst_file: os.PathLike, id_field: str
) -> None:
    src_file = Path(src_file).absolute().as_posix()
    dst_file = Path(dst_file).absolute().as_posix()
    fn_kwargs = {"src_file": src_file, "dst_file": dst_file, "id_field": id_field}
    call_script_fn(fn_name="_key_to_row_indices_mapping_op", kwargs=fn_kwargs)


def _group_qrel_triplets_from_csv_op(
    src_file: os.PathLike, dst_file: os.PathLike
) -> None:
    """Group query triplets from csv file using polars.

    ``src_file`` should contain rows of (qid, docid, score) triplets.
    The only restriction on the structure of the file (e.g., num columns, headers, etc.)
    is that it should be supported by ``csv_loader_kwargs()`` function and once
    loaded using its returned arguments, it must have three columns named ``qid``, ``docid``, and ``score``.

    The results are written to a json lines file where each row is a dict of following format::

        {
            'qid': ...,
            'docid': [docid1, docid2, ...], # ID of related documents
            'score': [score1, score2, ...]  # similarity scores of related documents in the same order.
        }

    Args:
        src_file (os.PathLike): a csv/tsv file to read qrel triplets from.
        dst_file (os.PathLike): a target jsonl file to write the grouped results to.
    """
    import polars as pl

    dst_file = Path(dst_file)
    if dst_file.suffix != ".jsonl":
        msg = f"We only support writing to jsonl file. But, got: {dst_file.as_posix()}"
        raise ValueError(msg)

    # polars needs the directory to exist
    dst_file.parent.mkdir(exist_ok=True, parents=True)

    _loader_kwargs = csv_loader_kwargs(src_file)
    pl_loader_kwargs = {
        "has_header": _loader_kwargs.get("header", None) is not None,
        "separator": _loader_kwargs["sep"],
        "new_columns": _loader_kwargs["names"],
    }
    # Scan instead of read so if possible all the data is not read into memory
    df = pl.scan_csv(src_file, **pl_loader_kwargs, cache=False)
    # group by 'qid' column and collect docids and scores for each group in a list
    df = df.group_by("qid").agg(pl.col("docid"), pl.col("score"))
    # Run the actual query
    df = df.collect()
    df.write_ndjson(dst_file.as_posix())


@functools.wraps(_group_qrel_triplets_from_csv_op)
def group_qrel_triplets_from_csv_op(
    src_file: os.PathLike, dst_file: os.PathLike
) -> None:
    src_file = Path(src_file).absolute().as_posix()
    dst_file = Path(dst_file).absolute().as_posix()
    fn_kwargs = {"src_file": src_file, "dst_file": dst_file}
    call_script_fn(fn_name="_group_qrel_triplets_from_csv_op", kwargs=fn_kwargs)


def _group_qrel_triplets_from_sydir_corpus_op(
    src_file: os.PathLike, dst_file: os.PathLike
) -> None:
    """Group inferred qrel triplets from sydir docids.

    sydir docids are formatted as ``f"{qid}_l_{level}_d_{doc_idx}"``.
    We parse this and use ``level`` as the ``score`` field in the qrel triplets.

    see ``group_qrel_triplets_from_csv()`` function for the format of the output file.

    Args:
        src_file (os.PathLike): corpus file containing sydir synthetic documents in jsonl format.
            Each record is a document with either ``docid`` or ``_id`` field that follows sydir docid formatting.
        dst_file (os.PathLike): a target jsonl file to write the grouped results to.
    """
    import polars as pl

    dst_file = Path(dst_file)
    if dst_file.suffix != ".jsonl":
        msg = f"We only support writing to jsonl file. But, got: {dst_file.as_posix()}"
        raise ValueError(msg)

    # Find out the field name that holds docid values
    with open(src_file, "r") as f:
        _rec = json.loads(f.readline().strip())
    if "_id" in _rec and "docid" in _rec:
        msg = "The records contain both 'docid' and '_id' field. I cannot choose which one to use."
        raise RuntimeError(msg)
    elif "_id" in _rec:
        id_field = "_id"
    elif "docid" in _rec:
        id_field = "docid"
    else:
        msg = f"I could not find either '_id' or 'docid' fields in the given file: {src_file}"
        raise RuntimeError(msg)

    # polars needs the directory to exist
    dst_file.parent.mkdir(exist_ok=True, parents=True)

    # Scan instead of read so if possible all the data is not read into memory
    df = pl.scan_ndjson(src_file)
    # We only 'docid' field to create qrel triplets
    df = df.select(pl.col(id_field).alias("docid"))

    # we want to stick to polars builtin operations and avoid python functions since they are much faster
    # But the current version of polars cannot split text from right (does not have something like str.rsplit)
    # So, we reverse the text, split it, take a specific split, and reverse it again
    df = df.with_columns(
        pl.col("docid")
        .str.reverse()
        .str.splitn("_", n=5)
        .struct.field("field_4")  # this is the qid split
        .str.reverse()
        .alias("qid"),
        pl.col("docid")
        .str.reverse()
        .str.splitn("_", n=4)
        .struct.field("field_2")  # this is the score (i.e., level) split
        .str.reverse()
        .cast(pl.Int32)  # Makes computation more efficient
        .alias("score"),
    )
    # group by 'qid' column and collect docids and scores for each group in a list
    df = df.group_by("qid").agg(pl.col("docid"), pl.col("score"))
    # Run the actual query
    df = df.collect()
    df.write_ndjson(dst_file.as_posix())


@functools.wraps(_group_qrel_triplets_from_sydir_corpus_op)
def group_qrel_triplets_from_sydir_corpus_op(
    src_file: os.PathLike, dst_file: os.PathLike
) -> None:
    src_file = Path(src_file).absolute().as_posix()
    dst_file = Path(dst_file).absolute().as_posix()
    fn_kwargs = {"src_file": src_file, "dst_file": dst_file}
    call_script_fn(
        fn_name="_group_qrel_triplets_from_sydir_corpus_op", kwargs=fn_kwargs
    )


if __name__ == "__main__":
    fn_name = sys.argv[1]
    fn_kwargs = sys.argv[2:]
    if any(["=" in a for a in fn_kwargs]):
        msg = f"Only '--key' 'value' format is supported. But it is possible that args of '--key=value' formats were given. args: {sys.argv}"
        raise ValueError(msg)

    fn_kwargs = [akv for a in fn_kwargs for akv in a.split("=", maxsplit=1)]
    fn_kwargs = {
        key.lstrip("-"): value for key, value in zip(fn_kwargs[::2], fn_kwargs[1::2])
    }
    if fn_name == "_key_to_row_indices_mapping_op":
        _key_to_row_indices_mapping_op(**fn_kwargs)
    elif fn_name == "_group_qrel_triplets_from_csv_op":
        _group_qrel_triplets_from_csv_op(**fn_kwargs)
    elif fn_name == "_group_qrel_triplets_from_sydir_corpus_op":
        _group_qrel_triplets_from_sydir_corpus_op(**fn_kwargs)
    else:
        msg = (
            f"Function '{fn_name}' is not recognized. Only supported functions are"
            " ['_key_to_row_indices_mapping_op', '_group_qrel_triplets_from_csv_op', '_group_qrel_triplets_from_sydir_corpus_op']."
        )
        raise ValueError(msg)
